"""For calling errors if a user tries run certain operations in the wrong python environment"""

from importlib import import_module
from typing import Union

package_name = 'blendersynth'

def conditional_import(condition: bool, src: str, name:str=None) -> Union['IllegalImport', object]:
	"""
	[No longer used in __init__.py, but kept here for reference]
	Imports a module if condition,
	otherwise returns IllegalImport.

	:param condition: If True, import the module
	:param src: The module to import
	:param name: The name of the module to import. If None, the last part of src is used.
	:return: The imported module, or IllegalImport if condition is False
	"""

	if condition:
		if src.startswith('.'):
			module = import_module(src, package=package_name)
		else:
			module = import_module(src)

		if name is None:
			return module

		return getattr(module, name if name else src.split('.')[-1])
	else:
		return IllegalImport()

class IllegalImport:
	"""Object which raises error on any interaction, to tell users they have used BlenderSynth incorrectly."""
	def __init__(self, REQUIRES_BLENDER=True):
		self.REQUIRES_BLENDER = REQUIRES_BLENDER

	def __call__(self, *args, **kwargs):
		raise ImportError(self._message)

	def __getattr__(self, name):
		raise ImportError(self._message)

	@property
	def _message(self):
		if self.REQUIRES_BLENDER:
			return "This object can only be used inside a Blender python environment. " \
				   "Make sure that either:" \
				   "\n- bsyn.run_this_script() is run on a previous line" \
				   "\n- This script is being run from another python script, using bsyn.execute_jobs."

		else:
			return "This object can only be used outside a Blender python environment. "


import bpy
from .utils import project_points
import mathutils
import numpy as np
from ..blender.mesh import Mesh
from typing import List, Union
from .annotation_handler import AnnotationHandler
from ..blender.camera import Camera


def _get_object_axes(object: Mesh, camera: bpy.types.Camera = None,
					 scene: bpy.types.Scene = None,
					 normalized=True,
					 invert_y=True) -> np.ndarray:
	"""Get the unit axes of an object in camera space.

	:param object: Mesh
	:param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	:param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)

	:return: (4 x 2) centroid + 3 pixel positions of ends of unit axes [XYZ] in XY
	"""

	if camera is None:
		camera = bpy.context.scene.camera

	if scene is None:
		scene = bpy.context.scene

	if len(object._meshes) > 1:
		raise NotImplementedError('Axes only supported for single meshes')

	centroid = np.array(object.centroid())
	verts = np.concatenate([centroid[None, :], centroid[None, :] + object.axes], axis=0)
	coords_2d = project_points(verts, scene, camera, invert_y=invert_y)

	if normalized:
		coords_2d[:, 0] /= scene.render.resolution_x
		coords_2d[:, 1] /= scene.render.resolution_y

	return coords_2d


def _get_objects_axes(objects: List[Mesh], camera: bpy.types.Camera = None,
					  scene: bpy.types.Scene = None,
					  normalized: bool = False, invert_y: bool = True) -> List[np.ndarray]:
	"""Run `get_axes` for multiple objects

	:param objects: List of Mesh objects
	:param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	:param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
	"""

	return [_get_object_axes(obj, camera=camera, scene=scene,
							 normalized=normalized, invert_y=invert_y) for obj in objects]


def get_axes(objects: List[Mesh], camera: Union[Camera, List[Camera]] = None,
			 scene: bpy.types.Scene = None, normalized: bool = False, invert_y: bool = True) -> AnnotationHandler:
	"""Project local axes for all objects into all cameras

	:param objects: List of Mesh objects
	:param camera: Either a single :class:`~blendersynth.blender.camera.Camera` object, or a list of them. If None, use bpy.context.scene.camera
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	:param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
	"""

	return_dict = {}

	if camera is None:
		camera = bpy.context.scene.camera

	if not isinstance(camera, list):
		camera = [camera]

	for cam in camera:
		return_dict[cam.name] = _get_objects_axes(objects, camera=cam, scene=scene, normalized=normalized,
												  invert_y=invert_y)

	return AnnotationHandler.from_annotations(return_dict, ann_type='axes')



import bpy
from .bsyn_object import BsynObject
from .nodes import tidy_tree


def _new_shader_node(node_tree, node_type):
	shader_node = node_tree.nodes.new(type=node_type)
	if node_type == 'ShaderNodeBsdfPrincipled':
		shader_node.inputs['Specular'].default_value = 0

	return shader_node


class Material(BsynObject):
	"""BlenderSynth Material class. Will always be a node material."""

	def __init__(self, name, shader_type='ShaderNodeBsdfPrincipled', mat=None):
		if mat is None:
			mat = bpy.data.materials.new(name)
			mat.use_nodes = True
			mat.node_tree.nodes.clear()
			self.shader = _new_shader_node(mat.node_tree, shader_type)
			self.output = mat.node_tree.nodes.new(type="ShaderNodeOutputMaterial")
			mat.node_tree.links.new(self.shader.outputs[0], self.output.inputs['Surface'])

		else:
			self.shader = None
			self.output = mat.node_tree.nodes['Material Output']

		mat.use_nodes = True
		self._object = mat

		self.node_tree = mat.node_tree
		self.nodes = self.node_tree.nodes
		self.links = self.node_tree.links

		self._image_nodes = {}  # dict of shader input socket : image node

		# set-up scaling
		self._add_scaling_node()
		self._scale = 1

		tidy_tree(self.node_tree)


	@classmethod
	def from_image(cls, image_loc, name="ImageMaterials", shader_type='ShaderNodeBsdfPrincipled',
				   use_global_scale=True) -> 'Material':
		"""
		Create a new material from an image texture.
		:param image_loc: Location of the image
		:param name: Name of the material
		:param shader_type: Shader type
		:param use_global_scale: Flag to scale the image texture node by the global scale
		:return: New Material instance
		"""
		mat = cls(name, shader_type=shader_type)  # new instance
		mat.add_source(image_loc, input_name='Base Color', use_global_scale=use_global_scale)

		return mat

	def add_source(self, image_loc:str, input_name:str='Base Color', use_global_scale:bool=True,
				   allow_duplicate:bool=False):
		"""
		Add an image texture node to the material node tree, and connect it to the specified input of the Shader.
		:param image_loc: Image location
		:param input_name: Input socket name of the Shader to connect the image texture node to
		:param use_global_scale: Flag to scale the image texture node by the global scale
		:param allow_duplicate: Flag to allow duplicate image texture nodes (by default, a node to the same input will overwrite any previous)
		:return:
		"""
		# Get the nodes in the material node tree
		nodes = self.obj.node_tree.nodes
		links = self.obj.node_tree.links

		# Find the Principled BSDF
		principled = next((node for node in nodes if node.type == 'BSDF_PRINCIPLED'), None)
		if not principled:
			raise ValueError("add_source requires a Principled BSDF node in the material node tree")

		if input_name in self._image_nodes and not allow_duplicate:
			tex_image_node = self._image_nodes[input_name]
		else:
			tex_image_node = nodes.new('ShaderNodeTexImage')

		tex_image_node.image = bpy.data.images.load(image_loc)
		self._image_nodes[input_name] = tex_image_node

		if input_name == 'Displacement':
			# Connect to the 'Displacement' socket of the output node
			links.new(tex_image_node.outputs['Color'], self.output.inputs['Displacement'])

		else:
			# Connect the image texture node to the specified input of the Principled BSDF
			links.new(tex_image_node.outputs['Color'], principled.inputs[input_name])

		if use_global_scale:
			links.new(self._texture_scaling_socket, tex_image_node.inputs['Vector'])

		if 'color' not in input_name.lower():
			tex_image_node.image.colorspace_settings.name = 'Non-Color'

		tidy_tree(self.node_tree)

	@classmethod
	def from_blender_material(cls, blender_mat: bpy.types.Material):
		if not isinstance(blender_mat, bpy.types.Material):
			raise ValueError("Invalid Blender material")

		# Create a new Material instance
		mat = cls(blender_mat.name, mat=blender_mat)

		# Find the shader and output nodes in the copied node tree
		mat.shader = next((node for node in mat.nodes if node.type == 'BSDF_PRINCIPLED'), None)
		mat.output = next((node for node in mat.nodes if node.type == 'OUTPUT_MATERIAL'), None)

		return mat

	def _add_scaling_node(self):
		"""Add scaling node to the material node tree"""

		self.scale_node = self.nodes.new('ShaderNodeValue')
		tex_coord_node = self.nodes.new('ShaderNodeTexCoord')

		mapping_node = self.nodes.new('ShaderNodeMapping')

		self.links.new(self.scale_node.outputs[0], mapping_node.inputs['Scale'])
		self.links.new(tex_coord_node.outputs['UV'], mapping_node.inputs['Vector'])
		self._texture_scaling_socket = mapping_node.outputs['Vector']

	def set_bdsf_property(self, key: str, value: float):
		"""Set the property of the BSDF node"""
		if self.shader.type != 'BSDF_PRINCIPLED':
			raise ValueError("Shader is not a Principled BSDF")

		if key not in self.shader.inputs:
			raise ValueError(f"Invalid BDSF shader property `{key}`")

		self.shader.inputs[key].default_value = value

	@property
	def scale(self):
		return self._scale

	@scale.setter
	def scale(self, value):
		self._scale = value
		self.scale_node.outputs[0].default_value = value

	@property
	def name(self):
		return self.obj.name


from typing import Union, List, Tuple, ForwardRef

try:
	from mathutils import Vector
except ImportError:
	Vector = 'mathutils.Vector'

import numpy as np

types_loc = 'blendersynth.utils.types'

VectorLike = Union[Vector, np.ndarray, List, Tuple]
"""A type hint that represents vector-like objects. Can be any of:

- :class:`mathutils.Vector`
- :class:`numpy.ndarray`
- :class:`list`
- :class:`tuple`
"""

VectorLikeOrScalar = Union[VectorLike, int, float]
"""A type hint that represents vector-like objects. Can be any of:

- :class:`~VectorLike`
- :class:`int`
- :class:`float`
"""

## Annotations
KeypointOrAxesAnnotation = List[np.ndarray]
"""Keypoint Annotations, an N x 2 ndarray per instance,
or Axes annotation, a 4 x 2 ndarray per instance"""

BboxAnnotation = List[VectorLike]
"""Bbox annotation, a 1D array per instance"""


# convert our type hint to Sphinx so it can be searched for in documentaton
def sphinxify_type_hint(py_type):
	if py_type is List:  # Handle built-in types
		return ":py:class:`typing.List`"
	if py_type is Tuple:
		return ":py:class:`typing.Tuple`"
	if py_type is int:
		return ":py:class:`int`"
	if py_type is float:
		return ":py:class:`float`"

	if isinstance(py_type, ForwardRef):
		return f":py:class:`~{py_type.__forward_arg__}`"

	elif isinstance(py_type, str):
		return f":py:class:`~{py_type}`"


	elif hasattr(py_type, "__origin__") and hasattr(py_type, "__args__"):
		origin = py_type.__origin__
		args = py_type.__args__
		sphinx_args = ", ".join(sphinxify_type_hint(arg) for arg in args)

		if origin is Union:
			origin_name = 'Union'
		else:
			origin_name = origin.__name__ if hasattr(origin, '__name__') else str(origin).split('.')[-1]

		return f":py:data:`~typing.{origin_name}`[{sphinx_args}]"


	else:
		return f":py:class:`~{py_type.__module__}.{py_type.__name__}`"


sphinx_mappings = {}  # dictionary of Sphinx auto type hint -> Type hint within this file
wrapper_mappings = {}  # dictionary of object -> type hint within this file
for k in ['VectorLike', 'VectorLikeOrScalar', 'KeypointOrAxesAnnotation', 'BboxAnnotation']:
	obj = globals()[k]
	key = sphinxify_type_hint(obj)
	value = f':class:`{k} <{types_loc}.{k}>`'
	sphinx_mappings[key] = value
	wrapper_mappings[obj] = value

# to avoid circular referencing, we also create some more type hints here (e.g. Mesh)
Mesh = 'blendersynth.blender.mesh.Mesh'

for name in ['Mesh']:
	obj = globals()[name]
	sphinx_mappings[sphinxify_type_hint(obj)] = f':class:`{name} <{obj}>`'


import bpy
from .bsyn_object import BsynObject

prims = {
	'circle': bpy.ops.curve.primitive_bezier_circle_add,
	'bezier': bpy.ops.curve.primitive_bezier_curve_add,
	'path': bpy.ops.curve.primitive_nurbs_path_add,
}

class Curve(BsynObject):
	"""Blender path object"""
	def __init__(self, path_type='circle', scale=1, location=(0, 0, 0), rotation=(0, 0, 0)):
		"""Create a new path:

		:param path_type: Type of path to create. One of 'circle', 'bezier', 'path' (straight line)"""

		self.path = None
		self._create_path(path_type)
		self._object = self.path

		self.scale = scale
		self.location = location
		self.rotation = rotation

	def _create_path(self, path_type):
		prims[path_type]()
		self.path = bpy.context.object



from collections import namedtuple
from copy import deepcopy
from ..utils import types

# docs-special-members: __init__

ANN_TYPES = ['bbox', 'keypoints', 'axes']
"""Valid annotation types"""


class Annotation:
	"""A class for storing annotations for a single image."""

	def __init__(self, bbox: types.BboxAnnotation = None, keypoints: types.KeypointOrAxesAnnotation = None,
				 axes: types.KeypointOrAxesAnnotation = None):
		"""

		:param bbox: Bounding box tuple per instance
		:param keypoints: [N x 2] array per instance
		:param axes: [4 x 2] array per instance
		"""
		self.bbox = bbox
		self.keypoints = keypoints
		self.axes = axes

	def __getitem__(self, item):
		return getattr(self, item)

	def _set(self, item, value) -> 'Annotation':
		"""In place set, returning self"""
		setattr(self, item, value)
		return self

	def set(self, item: str, value, inplace:bool=False) -> 'Annotation':
		"""Return a new Annotation, with item set to value.

		:param item: A key in :class:`ANN_TYPES`
		:param value: Value to set
		:param inplace: If True, set in place and return self
		"""
		if inplace:
			return self._set(item, value)

		return self.copy()._set(item, value)

	def copy(self) -> 'Annotation':
		"""Return a deep copy of this Annotation instance"""
		return deepcopy(self)

	def union(self, other: 'Annotation', priority='other') -> 'Annotation':
		"""Joins two annotations together, returning a new Annotation instance.

		:param other: Annotation to join with
		:param priority: Which annotation to take priority when merging
		"""

		combined = Annotation()

		low_priority = 'self' if priority == 'other' else 'other'

		for ann_type in ANN_TYPES:
			vals = {'self': self[ann_type], 'other': other[ann_type]}

			if vals[priority] is not None:
				combined._set(ann_type, vals[priority])

			elif vals[low_priority] is not None:
				combined._set(ann_type, vals[low_priority])

		return combined

	def __add__(self, other: 'Annotation'):
		return self.union(other)


class AnnotationHandler:
	"""An object for storing and handling per-view :class:`Annotation` instances"""

	def __init__(self, data: dict):
		"""
		:param data: Dictionary of camera_name: Annotation class
		"""

		self._data = data

	@classmethod
	def from_annotations(cls, data: dict, ann_type: str = 'bbox') -> 'AnnotationHandler':
		"""Create an Annotation Handler from a single type of annotation

		:param data: Dictionary of camera_name: annotation (e.g. bboxes)
		:param ann_type: Type of annotation (e.g. 'bbox')
		"""

		assert ann_type in ANN_TYPES, f'Invalid annotation type: {ann_type}. Must be one of {ANN_TYPES}'
		out_data = {}

		for cam_name, v in data.items():
			ann = Annotation(**{ann_type: v})
			out_data[cam_name] = ann

		return cls(out_data)

	def union(self, other: 'AnnotationHandler', priority: str = 'other') -> 'AnnotationHandler':
		"""Combine two AnnotationHandlers, merging annotations wherever possible. Returns a new instance
		comprising the union of the two AnnotationHandlers.

		:param other: AnnotationHandler to merge with
		:param priority: Which AnnotationHandler to take priority when merging
		"""

		assert isinstance(other, AnnotationHandler), f'other must be an AnnotationHandler, not {type(other)}'

		out_data = {}
		camera_names = set(self._data.keys()).union(set(other._data.keys()))

		for camera_name in camera_names:
			self_ann, other_ann = self.get_annotation_by_camera(camera_name), other.get_annotation_by_camera(
				camera_name)
			out_data[camera_name] = self_ann.union(other_ann, priority=priority)

		return AnnotationHandler(out_data)

	def get_annotation_by_camera(self, camera_name) -> Annotation:
		"""Return annotation for a given camera (default to an empty :class:`Annotation` instance if not found)"""
		return self._data.get(camera_name, Annotation())

	def __add__(self, other: 'AnnotationHandler'):
		return self.union(other)



"""Armature object for managing pose"""
import numpy as np

from .bsyn_object import BsynObject, animatable_property
from .other_objects import Empty
from .utils import SelectObjects, SetMode
from typing import Union
from ..utils import types
import bpy
import mathutils



class PoseBone(BsynObject):
	def __init__(self, object: bpy.types.PoseBone):
		self._object = object
		object.rotation_mode = 'XYZ'  # all rotations done in XYZ Euler mode


	@property
	def name(self) -> str:
		return self.obj.name

	@property
	def constraints(self):
		return self.obj.constraints

	@property
	def armature(self):
		return self.obj.id_data

	@property
	def matrix_world(self):
		return self.armature.matrix_world @ self.obj.matrix

	@property
	def location(self) -> mathutils.Vector:
		"""
		Gets the world location of the head (start) of the bone.
		"""
		return self.obj.head

	@property
	def head_location(self) -> mathutils.Vector:
		"""
		Gets the world location of the head (start) of the bone.
		"""
		return self.armature.matrix_world @ self.obj.head

	@property
	def tail_location(self) -> mathutils.Vector:
		"""
		Gets the world location of the tail (or end) of the bone.
		"""
		return self.armature.matrix_world @ self.obj.tail

	@property
	def rotation_euler(self) -> mathutils.Euler:
		"""Rotation of pose bone in world space"""
		return self.matrix_world.to_euler()

	@property
	def local_rotation_euler(self) -> mathutils.Euler:
		"""Rotation of pose bone in local space (ignoring parent transforms)"""
		return self.obj.rotation_euler

	@local_rotation_euler.setter
	def local_rotation_euler(self, value: mathutils.Euler):
		self.obj.rotation_euler = value

	def apply_global_rotation(self, axis:str='X', val:float=0.0, degrees=False):
		"""Equivalent to selecting the bone in Blender, and applying a global rotation about an axis

		:param axis: X, Y, or Z
		:param val: Rotation amount in radians
		:param degrees: If true, rotation is given in degrees"""

		if degrees:
			val = np.deg2rad(val)

		with SetMode('POSE', object=self.armature):
			with SelectObjects([self.obj]):
				bpy.context.object.data.bones.active = self.armature.data.bones[self.name]

				bpy.ops.transform.rotate(value=val, orient_axis=axis, orient_type = 'GLOBAL',
										 constraint_axis=[axis=='X', axis=='Y', axis=='Z'])

	@animatable_property('scale')
	def set_scale(self, scale: types.VectorLikeOrScalar, update:bool = True):
		"""Set scale of pose bone.

		:param scale: Scale to set. Either single value or 3 long vector
		:param update: Update the scene after setting
		scale. For batch operations, can be faster to set this to False, and call `bsyn.context.view_layer.update()`
		after all operations.
		"""

		if isinstance(scale, (int, float)):
			scale = (scale, scale, scale)

		self.obj.scale = scale

		if update:
			bpy.context.view_layer.update()


class BoneConstraint(BsynObject):
	"""Generic pose bone constraint object"""
	def __init__(self, pose_bone: PoseBone, constraint: bpy.types.Constraint, empty: Empty):
		self.pose_bone = pose_bone
		self.constraint = constraint
		self.empty = empty
		self._object = empty.obj


	@property
	def name(self):
		return f"{self.constraint.name}: {self.pose_bone.name} -> {self.empty.name}"

	@property
	def bone(self):
		return self.pose_bone

	def remove(self):
		"""Remove this constraint."""
		self.bone.constraints.remove(self.constraint)
		if self.empty is not None:
			self.empty.remove()

	@property
	def armature(self):
		return self.bone.id_data

class Armature(BsynObject):
	"""This class manages armatures - a collection of posable bones."""

	def __init__(self, object: bpy.types.Armature):
		self._object = object
		self.ik_constraints = {}  # for IK constraints
		self.constraints = {} # for generic constraints
		self.pose_bones = {n: PoseBone(b) for n, b in object.pose.bones.items()}

	@property
	def name(self) -> str:
		return self.obj.name

	@property
	def pose(self) -> bpy.types.Pose:
		return self.obj.pose

	def get_bone(self, bone: Union[str, PoseBone, bpy.types.PoseBone]) -> PoseBone:
		"""
		Get bone from armature.

		:param bone_name: Name of bone to get (or PoseBone object)
		:return: PoseBone object
		"""

		if isinstance(bone, PoseBone):
			return bone

		if isinstance(bone, bpy.types.PoseBone):
			return self.pose_bones[bone.name]

		if isinstance(bone, str):
			return self.pose_bones[bone]

		raise TypeError(f"Expected bone to be PoseBone, str, or PoseBone, got {type(bone)}")

	def pose_bone(self, bone: Union[str, bpy.types.PoseBone], rotation: types.VectorLike = None,
				  location: types.VectorLike = None, scale: types.VectorLike = None,
				  frame: int = None):

		"""Set the pose of a bone by giving a Euler XYZ rotation and/or location.

		:param bone: Name of bone to pose, or PoseBone object
		:param rotation: Euler XYZ rotation in radians
		:param location: Location of bone
		:param scale: Scale of bone
		:param frame: Frame to set pose on. If given, will insert keyframe here.
		"""

		with SetMode('POSE', object=self.obj):
			if isinstance(bone, str):
				bone = self.get_bone(bone)

			bone.rotation_mode = 'XYZ'
			if rotation is not None:
				bone.set_rotation_euler(rotation, frame=frame)

			if location is not None:
				bone.set_location(location, frame=frame)

			if scale is not None:
				bone.set_scale(scale, frame=frame)

	def clear_pose(self, rot=True, location=True, scale=True, bones=None):
		"""Clear the pose of the armature.
		For the target bones, sets poses to zero, and removes any IK constraints.

		:param rot: Clear rotation
		:param location: Clear location
		:param scale: Clear scale
		:param bones: List of bones to clear. If not given, will clear all bones.
		"""

		with SetMode('POSE', object=self.obj):
			for bone in self.pose.bones:
				if bones is None or bone.name in bones:
					if rot:
						bone.rotation_euler = (0, 0, 0)
					if location:
						bone.location = (0, 0, 0)
					if scale:
						bone.scale = (1, 1, 1)

				for bone_name in self.ik_constraints.keys():
					if bone_name == bone.name:
						self.ik_constraints[bone_name].remove()  # remove the constraint
						self.ik_constraints.pop(bone_name)  # remove reference to it

				for cname in [*self.constraints.keys()]:
					constraint = self.constraints[cname]
					if constraint.pose_bone.name == bone.name:
						self.constraints[cname].remove()
						self.constraints.pop(cname)


	def add_constraint(self, bone: Union[str, PoseBone], constraint_name: str, target:Empty=None, **kwargs) -> BoneConstraint:
		"""
		Applies a generic bone constraint. Returns BoneConstraint object

		:param bone: Bone to apply constraint to
		:param constraint_name:
		:param target: Empty target. If none given, will create one.

		Any other constraint specific keyword arguments given as well will be fed into the new constraint
		"""

		bone = self.get_bone(bone)

		# Add IK constraint to the specified bone
		constraint = bone.constraints.new(constraint_name)
		for k, v in kwargs.items():
			setattr(constraint, k, v)

		# Create an empty to serve as the IK target
		if target is None:
			target = Empty(name="Target_For_" + bone.name)

		constraint.target = target.object

		bpy.context.view_layer.update()

		constraint = BoneConstraint(bone, constraint, target)
		self.constraints[constraint.name] = constraint

		return constraint



"""Some preferential blender operations to run (e.g. setup) that will be run before debug mode is entered"""
import bpy
from ..blender import render


def on_script_open():
	"""
	On script open:

	- Set Engine to CYCLES
	- Delete default Cube & Light
	- Disable splash screen
	"""
	render.set_engine('CYCLES')

	bpy.data.objects['Cube'].select_set(True)
	bpy.data.objects['Light'].select_set(True)
	bpy.ops.object.delete()

	bpy.context.preferences.view.show_splash = False  # disable splash screen



"""Quick utility to run the current script from Blender"""
import subprocess
import inspect
import sys
import os
from ..utils.blender_setup.blender_locator import get_blender_path
from ..file.tempfiles import create_temp_file, cleanup_temp_files as cleanup
from shutil import copyfile


def _copy_over_script(filepath: str) -> str:
	"""Copies over a python script to a tempfile, returning the path.
	Removes certain lines so it can run in blender"""
	remove_lines_containing = ['.run_this_script']
	new_lines = []
	with open(filepath, 'r') as f:
		lines = f.readlines()
		for line in lines:
			if not any([s in line for s in remove_lines_containing]):
				new_lines.append(line)

	new_filepath = create_temp_file(ext='.py')
	with open(new_filepath, 'w') as f:
		f.writelines(new_lines)

	return new_filepath


def run_this_script(*args, open_blender: bool = False,
					debug: bool = False, IDE: str = 'PyCharm', port: int = 5678, host: str = 'localhost',
					**kwargs):
	"""Run the script in which this function is called from Blender.

	Will also place a copy of the script inside Blender.

	:param open_blender: If True, open a Blender instance after all code is executed, otherwise run in background
	:param debug: If True, will run in debug mode
	:param IDE: IDE to use for debugging. Currently only PyCharm and VSCode are supported
	:param port: Port to use for debugging
	:param host: Host to use for debugging

	args & kwargs are passed to the script being run as command line arguments.
	"""

	running_in_blender = 'bpy' in sys.modules

	caller_path = inspect.stack()[1].filename  # path of script that called this function

	if not running_in_blender:  # if blender is not running this script

		caller_dir = os.path.dirname(caller_path)
		env = os.environ.copy()
		env['PYTHONPATH'] = caller_dir + os.pathsep + env.get('PYTHONPATH', '')

		blender_path = get_blender_path()

		commands = [blender_path] + \
				   ['--background'] * (not open_blender) + \
				   ['--python', caller_path, '--']

		for arg in args:
			commands += [f'--{arg}']

		for key, value in kwargs.items():
			commands += [f'--{key}', str(value)]

		subprocess.call(commands, env=env)

		cleanup()  # cleanup temp files
		sys.exit()  # exit the script once blender is finished

	else:

		if debug:
			IDE = IDE.lower()
			assert IDE in ['pycharm', 'vscode'], f"IDE `{IDE}` not supported, only PyCharm and VSCode are supported"

			if IDE == 'pycharm':
				import pydevd_pycharm
				pydevd_pycharm.settrace(host, port=port, stdoutToServer=True, stderrToServer=True, suspend=False)

			elif IDE == 'vscode':
				import debugpy
				debugpy.listen((host, port))
				debugpy.wait_for_client()

		# blender is running this script
		if open_blender:
			# load the script into blender for viewing
			import bpy
			from ..utils import layout

			caller_path = bpy.path.abspath(caller_path)
			script_path = _copy_over_script(caller_path)

			text_block = bpy.data.texts.load(script_path)

			layout.change_area_to('DOPESHEET_EDITOR', 'TEXT_EDITOR')
			layout.get_area('TEXT_EDITOR').spaces[0].text = text_block



import sys
from tqdm import tqdm
from time import perf_counter, sleep
import numpy as np
from subprocess import Popen
import atexit

from datetime import datetime, timedelta
import os


def _list_split(list, chunks):
	"""Split a list into chunks"""
	return [[*x] for x in np.array_split(list, chunks)]


class BlenderThread():
	"""Manages a list of jobs, which it feeds into sequential Blender instances in a single thread."""
	def __init__(self, command, jobs, log_loc, progress_loc, name='', timeout: int = 100, to_stdout: bool = False,
				 MAX_PER_JOB: int = 100, script_directory: str = None):
		"""
		:param progress_loc: .log file to write progress to
		:param timeout: longest time in (s) without render after which process is finished/failed.
		:param to_stdout: If True, print to stdout instead of to a log file.
		:param MAX_PER_JOB: Split command into jobs of size MAX_PER_JOB, and run each job in a separate process.
		:param script_directory: If given, add this to `sys.path` before running the script.
		"""

		self.command = command
		self.jobs = _list_split(jobs, np.ceil(len(jobs) / MAX_PER_JOB))  # split jobs into chunks of MAX_PER_JOB
		self.njobs = len(self.jobs)

		self.size = len(jobs)
		self.timeout = timeout
		self.script_directory = script_directory

		self.prev_n = 0  # store how many rendered to check for updates for timeout purposes
		self.timer = perf_counter()

		self.to_stdout = to_stdout
		if not self.to_stdout and log_loc is not None:
			self.log_loc = log_loc
			self.logfile = open(self.log_loc, "a")
		else:
			self.log_loc, self.logfile = None, None

		self.logger_loc = progress_loc

		self.process = None
		self.job = -1
		self.finished = False

		self.name = name
		self.status = f'STARTED THREAD {self.name}'

	def check_in(self):
		"""Checks if current job still running, if not move to next job.
		If all jobs complete, set self.finished = True"""
		if self.is_running:
			return

		self.job += 1

		if self.job >= self.njobs:
			self.finished = True
			self.status = f'✓ THREAD {self.name} COMPLETED.'
			return

		self.status = f"THREAD {self.name} RUNNING JOB {self.job + 1}/{self.njobs}..."
		self.start_job(self.job)

	def start_job(self, job:int=0):
		"""
		:param job: Index of job to start
		"""
		job_list = self.jobs[job]
		command = self.command.set_job(job_list).set_logger(self.logger_loc).command

		stdout = sys.stdout if self.to_stdout else self.logfile
		stderr = sys.stderr if self.to_stdout else self.logfile

		env = os.environ.copy()
		if self.script_directory is not None:
			env['PYTHONPATH'] = self.script_directory + os.pathsep + env.get('PYTHONPATH', '')

		self.process = Popen(command, universal_newlines=True, stdout=stdout, stderr=stderr, env=env)

	def terminate(self):
		"""End process"""
		self.process.kill()

	def kill(self):
		self.terminate()

	@property
	def is_running(self):
		if self.process:
			return self.process.poll() is None
		return False

	@property
	def complete(self):
		if not self.is_running:
			self.logfile.flush()
			return True

	def __len__(self):
		return self.size

	@property
	def num_rendered(self):
		"""Read through Log file to find how many renders have been completed"""
		if self.process is None or not os.path.isfile(self.logger_loc):
			return 0  # process not started yet

		x = 0
		with open(self.logger_loc, 'r') as f:
			for line in f.readlines():
				x += 1  # currently only 1 type of message to logger - render complete

		return x

	def remaining_idxs(self):
		raise NotImplementedError()

	@property
	def success(self):
		return self.num_rendered == self.size

	def check_status(self):
		"""True if still running successfully, False if exceeded timeout"""
		n = self.num_rendered
		if n > self.prev_n:
			self.prev_n = n
			self.timer = perf_counter()

		elif (perf_counter() - self.timer) >= self.timeout:
			self.status = f'✖ THREAD {self.name} FAILED [TIMEOUT].'
			return False

		return True


class BlenderThreadManager:
	"""Manager of multiple :class:`~blendersynth.run.blender_threading.BlenderThread` instances"""

	def __init__(self, command, jsons, output_directory, print_to_stdout=False,
				 MAX_PER_JOB=100, script_directory=None):
		"""
		:param commands: Base Blender command to run
		:param jsons: A list of num_threads size, each element is a list of jsons to render from
		:param log_locs:
		:param MAX_PER_JOB: To prevent memory issues, split up jobs into chunks of MAX_PER_JOB
		:param script_directory: If given, add this to `sys.path` before running the script.
		"""
		self.num_threads = len(jsons)

		self.command = command

		# create logs
		session_name = datetime.now().strftime(r"%y%m%d-%H%M%S")
		log_dir = os.path.join(output_directory, 'logs', session_name)
		progress_dir = os.path.join(log_dir, '_progress')  # for storing progress

		os.makedirs(log_dir, exist_ok=True)
		os.makedirs(progress_dir, exist_ok=True)

		logs = [os.path.join(log_dir, f'log_{i:02d}.txt') for i in range(self.num_threads)]
		progresses = [os.path.join(progress_dir, f'progress_{i:02d}.log') for i in range(self.num_threads)]
		self.log_locs = logs

		# Set report name as report_xx, incrementing by 1 each report
		report_fname = f"report_{len([f for f in os.listdir(output_directory) if 'report' in f]):02d}.txt"
		self.report_loc = os.path.join(output_directory, report_fname)

		self.t0 = 0
		self.session_start = None

		self.threads = []
		for i in range(self.num_threads):
			thread = BlenderThread(command,
								   jobs=jsons[i],
								   log_loc=None if logs is None else logs[i],
								   progress_loc=progresses[i],
								   name=str(i),
								   to_stdout=print_to_stdout,
								   MAX_PER_JOB=MAX_PER_JOB,
								   script_directory=script_directory)

			self.threads.append(thread)

	def __len__(self):
		return sum(map(len, self.threads))

	def start(self, progress_bars: bool = True,
			  tick: float = 0.5, report_every: float = 15.0, offset: float = 1.0):
		"""Start all threads and job progress
		:param progress_bars: Show progress bars for each thread and overall progress
		:param tick: How often to update progress bars
		:param report_every: How often to print status updates to log
		:param offset: How long to wait before starting each thread (to avoid memory issues)
		"""
		self.t0 = perf_counter()  # log start time
		self.session_start = datetime.now()

		for thread in self.threads:
			thread.check_in()
			sleep(offset)

		# register atexit handler
		atexit.register(self.terminate)

		last_report_time = perf_counter()

		if progress_bars:
			pbars = []
			bar_format = '{l_bar}{bar:10}{r_bar}{bar:-10b}'
			for i, thread in enumerate(self.threads):
				p = tqdm(total=len(thread), initial=thread.num_rendered, bar_format=bar_format,
						 position=i + 1)
				pbars.append(p)

			with tqdm(total=len(self), bar_format=bar_format, position=0) as pbar:
				while any(t.is_running for t in self.threads):
					sleep(tick)

					# Update progress bar
					rendered_images = self.num_rendered

					desc = 'Session rendering...'
					desc += f' [Dataset: {rendered_images}/{len(self)}]'
					pbar.set_description(desc)

					pbar.n = self.num_rendered
					pbar.refresh()

					if (perf_counter() - last_report_time) >= report_every:
						self.update_report()
						last_report_time = perf_counter()

					# Update sub-progress bars, restart threads if needed
					for t, thread in enumerate(self.threads):

						if thread.finished:
							continue

						thread.check_in()

						if thread.success:
							pbars[t].n = thread.num_rendered

						else:
							thread_running = thread.check_status()

							if thread_running:
								pbars[t].n = thread.num_rendered

							else:  # Thread failed, currently will only notify the user
								pass

						pbars[t].set_description(thread.status)

			self.update_report()

	@property
	def num_rendered(self):
		return sum(t.num_rendered for t in self.threads)

	def update_report(self):

		t1 = perf_counter()
		elapsed = t1 - self.t0

		report = []
		if self.num_rendered > 0:
			# calculate number of seconds remaining
			s_remaining = (len(self) - self.num_rendered) * (elapsed / self.num_rendered)
			eta = datetime.now() + timedelta(seconds=s_remaining)

			report += [
				f"Number of images rendered: {self.num_rendered}\n",
				f"Total session quota: {len(self)}\n",
				f"Time elapsed: {timedelta(seconds=round(elapsed))}\n",
				f"Time per render (s): {elapsed / self.num_rendered:.2f}\n\n",

				f"Session start: {self.session_start.strftime('%I:%M %p %d/%m/%y')}\n"
				f"Estimated End: {eta.strftime('%I:%M %p %d/%m/%y')}"
			]

		with open(self.report_loc, 'w') as outfile:
			outfile.writelines(report)

	def terminate(self):
		"""End all threads"""
		for thread in self.threads:
			thread.terminate()


import bpy
import numpy as np
import mathutils
from .utils import handle_vec, SelectObjects, animatable_property
from .bsyn_object import BsynObject
from typing import Union
from .mesh import Mesh
from .curve import Curve
from ..utils import types


def _look_at_rotation(obj_camera: bpy.types.Object,
					  at: mathutils.Vector = mathutils.Vector((0, 0, 0)),
					  up: mathutils.Vector = mathutils.Vector((0, 0, 1))) -> mathutils.Euler:
	"""	Rotate camera to look at 'at', with 'up' maintained.

	:param obj_camera: Camera object
	:param at: Point to look at
	:param up: Up vector"""

	camera_position = obj_camera.location

	z_axis = - (at - camera_position).normalized()  # Blender forwards is '-z', so use -ve here
	x_axis = up.cross(z_axis)
	y_axis = z_axis.cross(x_axis).normalized()
	is_close = all(abs(i) < 5e-3 for i in x_axis)

	if is_close:
		replacement = y_axis.cross(z_axis).normalized()
		x_axis = replacement

	R = mathutils.Matrix([x_axis, y_axis, z_axis]).transposed()
	euler = R.to_euler()
	return euler


class Camera(BsynObject):
	"""Camera object, to handle movement, tracking, etc."""

	def __init__(self, camera: bpy.types.Object = None):
		if camera is None:
			camera = bpy.context.scene.camera

		self.camera = camera
		self._object = camera

	@classmethod
	def create(cls, name: str = 'Camera',
			   location: Union[list, tuple, np.ndarray] = None,
			   rotation: Union[list, tuple, np.ndarray] = None) -> 'Camera':
		"""Create a new camera object.

		:param name: Name of camera
		:param location: Location of camera
		:param rotation_euler: Rotation of camera in euler angles"""

		with SelectObjects():  # revert selection after creating camera
			bpy.ops.object.camera_add(enter_editmode=False, align='VIEW')
			camera = bpy.context.object
			camera.name = name

		cam = cls(camera)

		if location is not None:
			cam.set_location(location)

		if rotation is not None:
			cam.set_rotation_euler(rotation)

		return cam

	@classmethod
	def from_scene(cls, name: str = 'Camera') -> 'Camera':
		"""Create from a named camera in scene

		:param name: Name of camera"""
		assert name in bpy.data.objects, f'Camera {name} does not exist'
		return cls(bpy.data.objects[name])

	def update(self):
		"""Update view layer and depsgraph"""
		bpy.context.view_layer.update()
		bpy.context.evaluated_depsgraph_get()

	@property
	def fov(self):
		"""Field of view in degrees"""
		return self.camera.data.angle * 180 / np.pi

	@fov.setter
	def fov(self, fov):
		self.set_fov(fov)

	@animatable_property('lens', 'data')  # FOV is not animatable, so keyframe focal length instead
	def set_fov(self, fov):
		self.camera.data.angle = fov * np.pi / 180
		self.update()

	@property
	def clip_start(self):
		return self.camera.data.clip_start

	@clip_start.setter
	def clip_start(self, clip_start):
		self.set_clip_start(clip_start)

	@animatable_property('clip_start', 'data')
	def set_clip_start(self, clip_start):
		self.camera.data.clip_start = clip_start
		self.update()

	@property
	def clip_end(self):
		return self.camera.data.clip_end

	@clip_end.setter
	def clip_end(self, clip_end):
		self.set_clip_end(clip_end)

	@animatable_property('clip_end', 'data')
	def set_clip_end(self, clip_end):
		self.camera.data.clip_end = clip_end
		self.update()

	def look_at(self, at: types.VectorLike = mathutils.Vector((0, 0, 0)),
				up: types.VectorLike = mathutils.Vector((0, 0, 1))):
		"""Look at a point in space, with up vector.

		:param at: Point to look at
		:param up: Up vector"""
		at = handle_vec(at, 3)
		up = handle_vec(up, 3)

		self.rotation_euler = _look_at_rotation(self.camera, at, up)

	def look_at_object(self, obj: Union[Mesh, bpy.types.Object],
					   up: types.VectorLike = mathutils.Vector((0, 0, 1))):
		"""Look at an object, with up vector.

		:param obj: Object to look at
		:param up: Up vector"""

		if isinstance(obj, Mesh):
			obj = obj.obj

		self.look_at(obj.location, up)

	def place_and_rotate(self, pos, euler):
		self.location = pos
		self.euler = euler

	def place_and_look_at(self, pos, at, up=mathutils.Vector((0, 0, 1))):
		pos = handle_vec(pos)
		at = handle_vec(at)
		up = handle_vec(up)

		self.location = pos
		self.look_at(at, up)



# First, make sure Blender has been set up properly
import sys
from .utils.blender_setup import check_blender_install

def fix_blender_install(local=False, editable=False):
	check_blender_install(force_all=True, blendersynth_from_local=local, blendersynth_editable=editable)

def fix_blender_modules(local=False, editable=False):
	check_blender_install(force_install_dependencies=True, blendersynth_from_local=local,
						  blendersynth_editable=editable)

from .run.run_this_script import run_this_script

from .utils.blender_setup.blender_locator import get_blender_path
from .run.import_handling import conditional_import, IllegalImport
from typing import TYPE_CHECKING

is_building_docs = 'sphinx' in sys.modules
is_blender = 'bpy' in sys.modules
BLENDER_IMPORTS = is_building_docs or is_blender  # if blender is running this script, or if building docs

if BLENDER_IMPORTS or TYPE_CHECKING:
	from .blender.mesh import Mesh
	from .blender.material import Material
	from .blender.curve import Curve
	from .blender.other_objects import Empty
	from .blender import render
	from .blender.compositor.compositor import Compositor
	from .blender import aov
	from .file.dataset_inputs import Inputs, DebugInputs
	from .file.tempfiles import cleanup_temp_files
	from .run.blender_interface import log_event
	from .blender.world import world
	from .blender.light import Light
	from .blender.camera import Camera
	from . import annotations
	from .utils import layout
	from .blender import utils as blender_utils

	import mathutils
	import bpy

	# set IllegalImports here to warn users if they access the objects in the wrong python environment
	execute_jobs, install_module = [IllegalImport(REQUIRES_BLENDER=False) for _ in range(2)]

else:
	# Vanilla only imports
	from .run.run import execute_jobs
	from .utils.blender_setup.check_blender_install import install_module

	# set IllegalImports here to warn users if they access the objects in the wrong python environment
	Mesh, Material, Curve, Empty, render, Compositor, aov, Inputs, DebugInputs, cleanup_temp_files, log_event, world, \
	Light, Camera, annotations, layout, blender_utils, *___ = [IllegalImport() for _ in range(99)]



# Global imports
from . import file

# scripting setup
if is_blender:
	from bpy import *  # import everything from bpy

	def load_blend(src):
		return bpy.ops.wm.open_mainfile(filepath=src)

	# set render engine to cycles
	from .run.pre_ops import on_script_open
	on_script_open()

if not (BLENDER_IMPORTS or TYPE_CHECKING):
	check_blender_install(blendersynth_from_local='--local' in sys.argv)  # check install here


import os

from .blender_threading import BlenderThreadManager, _list_split
from ..utils.blender_setup.blender_locator import get_blender_path
from ..file.tempfiles import cleanup_temp_files as cleanup
from typing import Union

from sys import platform

if platform == "linux" or platform == "linux2" or platform == "darwin":
	nul_text = '/dev/null'
elif platform == "win32":
	nul_text = 'nul'


class BlenderCommand:
	"""Construct command for running blender script"""

	def __init__(self, blender_loc, background=True):
		self.blender_loc = blender_loc
		self.background = background
		self._command = [blender_loc, "--background" if background else ""]

	def compose(self, script, args=(), **kwargs):
		command = self._command + ["--python", script]
		command += ["--"]
		for arg in args:
			command.append(f"--{arg}")

		for k, v in kwargs.items():
			command += [f"--{k}", str(v)]

		self._command = command

	@property
	def command(self):
		return self._command

	def copy(self):
		"""Return a copy of the command"""
		cmd = BlenderCommand(self.blender_loc, self.background)
		cmd._command = [*self._command]
		return cmd

	def set_job(self, job_list):
		"""Receives list of json files, adds as , separated string to command"""
		cmd = self.copy()
		cmd._command += ["--jobs", ",".join(job_list)]
		return cmd

	def set_logger(self, log_loc):
		"""Set file for logging"""
		cmd = self.copy()
		cmd._command += ["--log", log_loc]
		return cmd


class Runner:
	def __init__(self, script, jsons, output_directory, num_threads=1,
				 print_to_stdout=False, distributed:tuple=None,
				 **script_kwargs):
		"""
		:param jsons: N sized list of .json files, each with info about the given job
		:param num_threads: threads to run in parallel (default = 1)
		:param distributed: tuple of (machine index, total machines) for distributed rendering
		"""

		# if distributed, split jsons into chunks
		if distributed is not None:
			machine_index, total_machines = distributed
			assert machine_index < total_machines, "machine_index must be less than total_machines"
			jsons = _list_split(jsons, total_machines)[machine_index]

		self.jsons = jsons
		self.num_threads = num_threads

		# Split the jsons into num_threads chunks
		json_chunks = _list_split(self.jsons, self.num_threads)

		blender_loc = get_blender_path()
		command = BlenderCommand(blender_loc=blender_loc, background=True)
		command.compose(script=script, **script_kwargs)

		thread_manager = BlenderThreadManager(command, json_chunks, print_to_stdout=print_to_stdout,
											  output_directory=output_directory,
											  script_directory=os.path.dirname(script))

		thread_manager.start()


def execute_jobs(script: str, json_src: Union[list, str], output_directory: str, num_threads: int = 1,
				 print_to_stdout: bool = False, distributed:tuple=None, **script_kwargs):
	"""
	Execute a script, given a list of jobs to execute.

	:param script: path to script to execute
	:param json_src: N sized list of .json files, each with info about the given job. OR a directory containing .json files,
		which will be used as the jsons
	:param output_directory: directory to save output files
	:param num_threads: threads to run in parallel (default = 1)
	:param print_to_stdout: If True, print to stdout instead of saving to file
	:param distributed: tuple of (machine index, total machines) for distributed rendering
	"""

	assert not (num_threads > 1 and print_to_stdout), "Cannot print to stdout with multiple threads"
	if print_to_stdout:
		raise NotImplementedError("Print to stdout not supported yet.")

	if isinstance(json_src, str):
		json_src = sorted([os.path.join(json_src, f) for f in os.listdir(json_src) if f.endswith(".json")])

	Runner(script, json_src, output_directory, num_threads, print_to_stdout=print_to_stdout,
			distributed=distributed,
		   **script_kwargs)

	cleanup()



import os.path

import bpy
from .utils import GetNewObject, SelectObjects, handle_vec, CursorAt, animatable_property
from .bsyn_object import BsynObject
from .material import Material
from .armature import Armature
from .aov import AOV
import numpy as np
import mathutils
import bmesh
from mathutils import Vector, Euler
from typing import Union, List
from copy import deepcopy
from ..utils import types

_primitives = {
	"cube": bpy.ops.mesh.primitive_cube_add,
	"sphere": bpy.ops.mesh.primitive_uv_sphere_add,
	"cylinder": bpy.ops.mesh.primitive_cylinder_add,
	"plane": bpy.ops.mesh.primitive_plane_add,
	"cone": bpy.ops.mesh.primitive_cone_add,
	"monkey": bpy.ops.mesh.primitive_monkey_add,
	"torus": bpy.ops.mesh.primitive_torus_add,
}

default_ids = {
	'loaded_mesh': 0,
	'prim_cube': 1,
	'prim_sphere': 2,
	'prim_cylinder': 3,
	'prim_plane': 4,
	'prim_cone': 5,
	'prim_monkey': 6,
	'prim_torus': 7,
}


def _get_child_meshes(obj):
	"""Given an object, return all meshes that are children of it. Recursively searches children of children.
	Also returns any child objects that aren't meshes"""
	if obj.type == 'MESH':
		return [obj], []
	else:
		meshes, other = [], [obj]
		for child in obj.children:
			child_meshes, child_other = _get_child_meshes(child)
			meshes += child_meshes
			other += child_other

		return meshes, other


def _set_object_origin(obj: bpy.types.Object, origin: Vector):
	"""Set the origin of an object to a given point"""
	with SelectObjects([obj]), CursorAt(origin):
		bpy.ops.object.origin_set(type='ORIGIN_CURSOR', center='MEDIAN')

def _find_material(obj):
	"""Return first material found in an object"""
	if obj.type == 'MESH':
		for mat in obj.data.materials:
			return mat

	for child in obj.children:
		if child.type == 'MESH':
			for mat in child.data.materials:
				return mat


class Mesh(BsynObject):
	"""A mesh object. Can be a single mesh, or a hierarchy of meshes."""
	primitive_list = list(_primitives.keys())
	"""List of available primitives"""

	def __init__(self, obj, material: bpy.types.Material = None, scene: bpy.types.Scene = None, class_id: int = None):
		"""
		:param obj: Receives either a single mesh, or an empty with children empty & meshes
		:param material: bpy.types.Material to assign to the mesh
		:param scene:

		If obj contains multiple meshes, they will all be assigned the same material, and must act as a single object
		i.e. they will all be transformed together

		"""

		if scene is None:
			scene = bpy.context.scene

		self.scene = scene
		self._object = obj
		self._meshes, self._other_objects = _get_child_meshes(obj)

		# Manage materials here
		if material is not None:
			self._material = Material.from_blender_material(material)
		elif _find_material(obj) is not None:
			self._material = Material.from_blender_material(_find_material(obj))
		else:
			self._material = Material('NewMaterial')

		for mesh in self._meshes:
			mesh.data.materials.append(self._material.object)

		# INSTANCING - Define InstanceID based on number of meshes in scene
		self._object['instance_id'] = scene.get('NUM_MESHES', 0)
		scene['NUM_MESHES'] = scene.get('NUM_MESHES', 0) + 1  # Increment number of meshes in scene

		# CLASS - Define class based on type of object (e.g. primitive)
		# can be overriden at any point
		self.set_class_id(class_id)

		self.assigned_aovs = []
		self._armatures = {}  # name : armature, to prevent multiple instances of armatures

	def set_class_id(self, class_id: int):
		"""Set the class ID of the object.

		:param class_id: Class ID to assign to object
		"""

		assert isinstance(class_id, int), f"Class ID must be an integer, not {type(class_id)}"
		assert class_id >= 0, f"Class ID must be >= 0, not {class_id}"

		self._object['class_id'] = class_id
		self.scene['MAX_CLASSES'] = max(self.scene.get('MAX_CLASSES', 0), class_id)

	@classmethod
	def from_scene(cls, key: str, class_id: int = 0) -> 'Mesh':
		"""Create object from named object in scene.

		:param key: Name of object in scene
		:param class_id: Class ID to assign to object

		:return: Mesh loaded from scene"""
		obj = bpy.data.objects[key]
		return cls(obj, class_id=class_id)

	@classmethod
	def from_primitive(cls, name='cube', scale=None, location=None, rotation_euler=None, class_id=None,
					   **kwargs) -> 'Mesh':
		"""Create Mesh from primitive.
		
		:param name: Name of primitive to create. See :attr:`~blendersynth.blender.Mesh.primitive_list` for options
		:param scale: Scale of object
		:param location: Location of object
		:param rotation_euler: Rotation of object
		:param class_id: Class ID to assign to object
		:param kwargs: Additional arguments to pass to primitive (see `bpy.ops.mesh.primitive_cube_add <https://docs.blender.org/api/current/bpy.ops.mesh.html#bpy.ops.mesh.primitive_cube_add>`_, etc.)

		:return: Mesh object created from primitive
		"""

		assert name in _primitives, f"Primitive `{name}` not found. Options are: {list(_primitives.keys())}"

		importer = GetNewObject(bpy.context.scene)
		with importer:
			prim = _primitives[name](**kwargs)  # Create primitive

		if class_id is None:
			class_id = default_ids[f'prim_{name}']

		obj = cls(importer.imported_obj, class_id=class_id)

		if scale is not None: obj.scale = scale
		if location is not None: obj.location = location
		if rotation_euler is not None: obj.rotation_euler = rotation_euler

		return obj

	@classmethod
	def from_obj(cls, obj_loc: str, class_id: int = None,
				 forward_axis: str = '-Z', up_axis: str = 'Y'):
		"""Load object from .obj file.

		:param obj_loc: Location of .obj file
		:param class_id: Class ID to assign to object
		:param forward_axis: Axis to use as forward axis
		:param up_axis: Axis to use as up axis


		Note: we use bpy.ops.wm.obj_import instead of bpy.ops.import_scene.obj because the latter
		causes issues with materials & vertex ordering.
		(Changing vertex ordering makes the use of keypoints difficult.)
		"""
		for axis in (forward_axis, up_axis):
			assert axis in (
				'X', 'Y', 'Z', '-X', '-Y', '-Z'), f"Axis `{axis}` not valid, must be one of X, Y, Z, -X, -Y, -Z"

		forward_axis = forward_axis.replace('-', 'NEGATIVE_')  # e.g. -X -> NEGATIVE_X
		up_axis = up_axis.replace('-', 'NEGATIVE_')

		assert os.path.isfile(obj_loc) and obj_loc.endswith('.obj'), f"File `{obj_loc}` not a valid .obj file"

		directory, fname = os.path.split(obj_loc)
		directory = os.path.abspath(directory)  # ensure absolute path for clean loading

		importer = GetNewObject(bpy.context.scene)
		with importer:
			bpy.ops.wm.obj_import(filepath=fname, directory=directory, filter_image=False,
								  files=[{"name": fname}], forward_axis=forward_axis, up_axis=up_axis)

		if class_id is None:
			class_id = default_ids['loaded_mesh']

		obj = importer.imported_obj
		return cls(obj, class_id=class_id)

	@classmethod
	def from_glb(cls, glb_loc: str, class_id: int = None):
		"""Load object from .glb file.

		:param glb_loc: Location of .glb file
		:param class_id: Class ID to assign to object
		"""
		assert os.path.isfile(glb_loc) and glb_loc.endswith(
			('.glb', '.gtlf')), f"File `{glb_loc}` not a valid .glb file"

		directory, fname = os.path.split(glb_loc)
		importer = GetNewObject(bpy.context.scene)
		with importer:
			bpy.ops.import_scene.gltf(filepath=glb_loc, files=[{"name": fname}])

		if class_id is None:
			class_id = default_ids['loaded_mesh']

		return cls(importer.imported_obj, class_id=class_id)

	@classmethod
	def from_gltf(cls, gltf_loc: str, class_id: int = None):
		"""Alias for :meth:`~blendersynth.blender.Mesh.from_glb`"""
		return cls.from_glb(gltf_loc, class_id=class_id)

	@classmethod
	def from_fbx(cls, fbx_loc: str, class_id: int = None):
		"""Load object from .fbx file.

		:param fbx_loc: Location of .fbx file
		:param class_id: Class ID to assign to object"""
		assert os.path.isfile(fbx_loc) and fbx_loc.endswith('.fbx'), f"File `{fbx_loc}` not a valid .fbx file"

		directory, fname = os.path.split(fbx_loc)
		importer = GetNewObject(bpy.context.scene)
		with importer:
			bpy.ops.import_scene.fbx(filepath=fbx_loc, files=[{"name": fname}])

		if class_id is None:
			class_id = default_ids['loaded_mesh']

		return cls(importer.imported_obj, class_id=class_id)

	@property
	def origin(self) -> Union[Vector, List[Vector]]:
		"""
		If single object, return Vector of origin.
		If multiple objects, return list of Vectors of centroid of each mesh."""
		if len(self._all_objects) == 1:
			return self._meshes[0].location

		else:
			return [m.location for m in self._all_objects]

	@origin.setter
	def origin(self, origin):
		"""Set origin of object.
		If single mesh, expects Vector.
		If multiple meshes, expects list of Vectors"""

		if len(self._all_objects) == 1:
			try:
				vec = handle_vec(origin)
			except ValueError:
				raise ValueError(f"Error with setting origin. Expects a 3 long Vector. Received: {origin}")

			_set_object_origin(self._meshes[0], vec)

		else:
			for i in range(len(self._all_objects)):
				try:
					vec = handle_vec(origin[i])
					_set_object_origin(self._all_objects[i], vec)
				except:
					raise ValueError(
						f"Error with setting origin. Expects a list of {len(self._all_objects)} 3-long Vector. Received: {origin}")

	def _get_all_vertices(self, ref_frame='WORLD') -> np.ndarray:
		"""
		Get all vertices of object, taking into account deformations.
		:param ref_frame: LOCAL or WORLD
		:return: Nx3 array of vertices
		"""

		depsgraph = bpy.context.evaluated_depsgraph_get()  # to account for deformations

		if ref_frame not in {'LOCAL', 'WORLD'}:
			raise ValueError(f"Invalid ref_frame: {ref_frame}. Must be one of ['LOCAL', 'WORLD']")

		verts = []

		for mesh in self._meshes:

			# use bmesh to get vertices - this accounts for deformations in depsgraph
			bm = bmesh.new()
			bm.from_object(mesh, depsgraph)
			bm.verts.ensure_lookup_table()
			mesh_verts = np.array([x.co for x in bm.verts])
			bm.free()

			if ref_frame == 'WORLD':
				mesh_verts = np.dot(mesh.matrix_world, np.vstack((mesh_verts.T, np.ones(mesh_verts.shape[0]))))

			verts.append(mesh_verts)

		verts = np.concatenate(verts, axis=1)
		verts /= verts[3]  # convert from homogeneous coordinates
		return verts[:3].T

	def get_raw_bounds(self) -> [Vector, Vector]:
		"""Get the minimum and maximum bounds of all meshes in the object, under no deformations or transformations."""
		verts = np.array([v.co for mesh in self._meshes for v in mesh.data.vertices])
		bbox_min = Vector([*np.min(verts, axis=0)])
		bbox_max = Vector([*np.max(verts, axis=0)])
		return bbox_min, bbox_max

	@property
	def materials(self):
		return {m for mesh in self._meshes for m in mesh.data.materials}

	def assign_pass_index(self, index: int):
		"""Assign pass index to object. This can be used when mask rendering.

		:param index: Pass index to assign"""
		for mesh in self._meshes:
			mesh.pass_index = index
		return index

	def assign_aov(self, aov: AOV):
		"""Assign AOV to object. Applies to all materials.

		:param aov: AOV to assign"""
		if aov not in self.assigned_aovs:
			for material in self.materials:
				shader_node_tree = material.node_tree
				assert shader_node_tree is not None, "Material must have a node tree"
				aov.add_to_shader(shader_node_tree)

		self.assigned_aovs.append(aov)

	def assign_aovs(self, aovs: List[AOV]):
		"""Assign multiple AOVs to object. Applies to all materials.

		:param aovs: AOVs to assign"""
		for aov in aovs:
			self.assign_aov(aov)

	def set_minimum_to(self, axis: str = 'Z', pos: float = 0):
		"""Set minimum of object to a given position in a given axis.

		:param axis: Axis to set minimum in
		:param pos: Position to set minimum to
		"""
		min_pos = self._get_all_vertices('WORLD')[:, 'XYZ'.index(axis)].min()
		trans_vec = np.zeros(3)
		trans_vec['XYZ'.index(axis)] = pos - min_pos
		self.translate(trans_vec)

	def clamp_in_axis(self, axis: str = 'Z', mode: str = 'min', value: float = 0) -> np.ndarray:
		"""Clamp object in a given axis to a given value - ensure that, in this axis,
		the object never goes below (mode='min') or above (mode='max') the given value.

		This can be used to ensure that an object never goes below the ground plane, for example.

		:param axis: Axis to clamp in
		:param mode: 'min' or 'max'
		:param value: Value to clamp to
		:return: Vector of translation applied to object
		"""

		axis = axis.upper()
		assert axis in 'XYZ', f"Invalid axis: {axis}"
		assert mode in {'min', 'max'}, f"Invalid mode: {mode}"

		verts = self._get_all_vertices('WORLD')
		axis_idx = 'XYZ'.index(axis)

		translation = 0
		if mode == 'min':
			translation = max(value - verts[:, axis_idx].min(), 0)  # can only be positive
		elif mode == 'max':
			translation = min(value - verts[:, axis_idx].max(), 0)  # can only be negative

		trans_vec = np.zeros(3)
		trans_vec['XYZ'.index(axis)] = translation
		self.translate(trans_vec)

		return trans_vec

	def delete(self, delete_materials: bool = True):
		"""Clear mesh from scene & mesh data.

		:param delete_materials: Also delete object materials from scene"""
		mesh_names = [m.name for m in self._meshes]
		for mesh in self._meshes:
			bpy.data.objects.remove(mesh, do_unlink=True)

		# remove any non-mesh objects
		for obj in self._other_objects:
			bpy.data.objects.remove(obj, do_unlink=True)

		# Also remove its mesh data from bpy.data.meshes
		for mname in mesh_names:
			for m in bpy.data.meshes:
				if m.name == mname:
					bpy.data.meshes.remove(m)
					break

		self._meshes = []  # clear mesh list

		if delete_materials:
			for material in self.materials:
				bpy.data.materials.remove(material, do_unlink=True)

	def centroid(self, method: str = 'median') -> Vector:
		"""
		Return the centroid of the mesh(es)

		:param method: See :attr:`~blendersynth.blender.mesh.Mesh.origin_to_centroid` for options.
		:return: Centroid of the mesh(es). If multiple meshes, will average the centroids.
		"""

		original_origins = deepcopy(self.origin)
		self.origin_to_centroid(method=method)
		centroid = deepcopy(self.origin)
		self.origin = original_origins

		if len(self._meshes) > 1:
			return sum(centroid) / len(self._meshes)

		return centroid

	def _set_origin_manual(self, origin: Vector, all_meshes=True):
		"""Override to set origin manually. Should only be used by internal functions."""
		if all_meshes:
			for mesh in self._meshes:
				mesh.location = origin
		else:
			self._meshes[0].location = origin

	def origin_to_centroid(self, method: str = 'bounds'):
		"""Move object origin to centroid.

		Four methods available:

		* 'bounds' - move the origin to the center of the bounds of the mesh
		* 'median' - move the origin to the median point of the mesh
		* 'com_volume' - move to the centre of mass of the volume of the mesh
		* 'com_area' - Move to the centre of mass of the surface of the mesh

		:param method: Selected method to move origin to centroid
		"""

		_valid_methods = ['bounds', 'median', 'com_volume', 'com_area']

		_type_lookup = dict(bounds='ORIGIN_GEOMETRY', median='ORIGIN_GEOMETRY',
							com_volume='ORIGIN_CENTER_OF_VOLUME',
							com_area='ORIGIN_CENTER_OF_MASS')

		center = 'BOUNDS' if method == 'bounds' else 'MEDIAN'

		with SelectObjects(self._meshes + self._other_objects):
			bpy.ops.object.origin_set(type=_type_lookup[method], center=center)

	def get_keypoints(self, idxs: list = None, position: Union[np.ndarray, List[Vector]] = None) -> List[Vector]:
		"""Return 3D keypoint positions in world coordinates, given either:

		:param idxs: list of indices or ndarray of keypoints to project (only valid for single-mesh objects)
		:param position: 3D position of keypoints to project - in LOCAL object coordinates

		:return: N list of Vectors of keypoints in world space, where N is the number of keypoints
		"""

		assert (idxs is not None) ^ (position is not None), "Must provide either idxs or position, but not both."

		if idxs is not None:
			assert len(self._meshes) == 1, "Can only project keypoints by index for single-mesh objects."
			kps3d = [self.matrix_world @ self._meshes[0].data.vertices[i].co for i in idxs]

		elif position is not None:
			kps3d = [self.matrix_world @ p for p in position]

		return kps3d

	@property
	def name(self):
		return self._meshes[0].name

	def get_armature(self, armature_name: str = None) -> Armature:
		"""Get armature.
		If no name given, return first armature found.

		:param armature_name: Name of armature to load."""

		armatures = [obj for obj in self._other_objects if obj.type == 'ARMATURE']

		if len(armatures) == 0:
			raise ValueError("No armatures found.")

		if armature_name is None:
			armature_name = armatures[0].name

		if armature_name in self._armatures:
			return self._armatures[armature_name]

		for arm in armatures:
			if arm.name == armature_name:
				bsyn_arm = Armature(arm)
				self._armatures[armature_name] = bsyn_arm
				return bsyn_arm

		raise KeyError(f"Armature `{armature_name}` not found.")

	@property
	def material(self):
		return self._material

	@material.setter
	def material(self, material: Material):
		self._material = material
		for mesh in self._meshes:
			mesh.data.materials.clear()
			mesh.data.materials.append(material.object)

	@property
	def shape_keys(self) -> bpy.types.ShapeKey:
		"""Get the first shape keys object available"""
		for mesh in self._meshes:
			if mesh.data.shape_keys is not None:
				return mesh.data.shape_keys

		raise ValueError("No mesh found with shape keys.")

	def set_shape_key(self, name: str, value: float, frame: int = None):
		"""Set shape key to a given value.

		:param name: Name of shape key
		:param value: Value to set shape key to
		:param frame: If not None, set keyframe at this frame
		"""

		shape_key = self.shape_keys.key_blocks[name]
		shape_key.value = value

		if frame is not None:
			shape_key.keyframe_insert(data_path="value", frame=frame)

	def set_shape_keys(self, data: dict, frame: int = None):
		"""Set multiple shape keys at once.

		:param data: Dictionary of shape key names to values
		:param frame: If not None, set keyframe at this frame
		"""
		for k, v in data.items():
			self.set_shape_key(k, v, frame=frame)



"""Custom node groups"""
import bpy
from .node_arranger import tidy_tree
import numpy as np
import os
import cv2

# docs-special-members: __init__
# no-inherited-members


class NodeGroup:
	"""Generic Node Group"""
	TYPE = 'Compositor'

	def __init__(self, name: str, node_tree: bpy.types.NodeTree):
		"""
		A generic NodeGroup class

		:param name: Name of node group
		:param node_tree: NodeTree to add group to
		"""
		self.name = name
		self.node_tree = node_tree
		self.group = bpy.data.node_groups.new(type=f'{self.TYPE}NodeTree', name=name)

		self.gn = group_node = node_tree.nodes.new(f"{self.TYPE}NodeGroup")
		group_node.node_tree = self.group

		self.input_node = self.group.nodes.new("NodeGroupInput")
		self.output_node = self.group.nodes.new("NodeGroupOutput")

	def tidy(self):
		tidy_tree(self.group)

	@property
	def inputs(self) -> dict:
		"""Input sockets"""
		return self.gn.inputs

	@property
	def outputs(self) -> dict:
		"""Output sockets"""
		return self.gn.outputs

	def input(self, name: str) -> bpy.types.NodeSocket:
		"""Get input socket by name"""
		return self.inputs[name]

	def output(self, name: str) -> bpy.types.NodeSocket:
		"""Get output socket by name"""
		return self.outputs[name]

	def add_node(self, key: str) -> bpy.types.Node:
		"""Create a new node in the group by name"""
		return self.group.nodes.new(key)

	def link(self, from_socket: bpy.types.NodeSocket, to_socket: bpy.types.NodeSocket) -> bpy.types.NodeLink:
		"""
		Link two sockets in the group

		:param from_socket: Socket to link from
		:param to_socket: Socket to link to
		"""
		return self.group.links.new(from_socket, to_socket)

	def __str__(self):
		return f"{self.TYPE}NodeGroup({self.name})"

	def update(self, camera=None, scene=None):
		pass

	def save_image(self, loc: str, image_data: np.ndarray):
		"""Save an image to a location. Will also reload any node reference to the image.

		:param loc: Location to save image to"""
		cv2.imwrite(loc, image_data)
		fname = os.path.basename(loc)
		if fname in bpy.data.images:
			bpy.data.images[fname].reload()

class CompositorNodeGroup(NodeGroup):
	"""Node Group for use in the compositor"""
	TYPE = 'Compositor'


class ShaderNodeGroup(NodeGroup):
	"""Node Group for use in the shader editor"""
	TYPE = 'Shader'



# # Add mask node & Mix node
# mask_node = self.get_mask(mask_index)
# mix_node = self.node_tree.nodes.new('CompositorNodeMixRGB')
# mix_node.inputs[1].default_value = (0, 0, 0, 0)  # black - 0 alpha
#
# self.node_tree.links.new(mask_node.outputs['Alpha'], mix_node.inputs['Fac'])
# self.node_tree.links.new(self.render_layers_node.outputs[input_name], mix_node.inputs[2])
# self.node_tree.links.new(mix_node.outputs['Image'], node.inputs['Image'])

from ..nodes import CompositorNodeGroup

class MaskOverlay(CompositorNodeGroup):
	def __init__(self, name, node_tree, index=0, use_antialiasing=True, dtype='Color'):
		"""
		Takes an IndexOB and an Image, and sets the values of all pixels with index != IndexOB
		to (0, 0, 0, 0)
		"""

		super().__init__(name, node_tree)

		# define I/O
		self.group.inputs.new('NodeSocketFloat', 'IndexOB')
		self.group.inputs.new(f'NodeSocket{dtype}', 'Image')
		self.group.outputs.new(f'NodeSocket{dtype}', 'Image')


		# create nodes
		self.mask_node = self.group.nodes.new('CompositorNodeIDMask')
		self.mask_node.index = index
		self.mask_node.use_antialiasing = use_antialiasing

		self.mix_node = self.group.nodes.new('CompositorNodeMixRGB')
		self.mix_node.inputs[1].default_value = (0, 0, 0, 1)  # to set color to black

		self.alpha_node = self.group.nodes.new('CompositorNodeSetAlpha')  # to set alpha to 0
		self.alpha_node.mode = 'APPLY'


		# link up internal nodes
		self.group.links.new(self.input_node.outputs['IndexOB'], self.mask_node.inputs['ID value'])
		self.group.links.new(self.input_node.outputs['Image'], self.mix_node.inputs[2])
		self.group.links.new(self.mask_node.outputs['Alpha'], self.mix_node.inputs['Fac'])
		self.group.links.new(self.mask_node.outputs['Alpha'], self.alpha_node.inputs['Alpha'])
		self.group.links.new(self.mix_node.outputs['Image'], self.alpha_node.inputs['Image'])
		self.group.links.new(self.alpha_node.outputs['Image'], self.output_node.inputs['Image'])




import bpy
from ..utils import get_node_by_name
import os
import shutil
from ..render import render, render_depth
from ..nodes import CompositorNodeGroup
from ..aov import AOV
from .mask_overlay import MaskOverlay
from .visuals import DepthVis
from .image_overlay import KeypointsOverlay, BoundingBoxOverlay, AlphaImageOverlay, AxesOverlay
from ...annotations import Annotation, AnnotationHandler
from ..nodes import tidy_tree
from ..world import world
from ..camera import Camera

from typing import Union, List

# Mapping of file formats to extensions
format_to_extension = {
	'BMP': '.bmp',
	'IRIS': '.rgb',
	'PNG': '.png',
	'JPEG': '.jpg',
	'JPEG2000': '.jp2',
	'TARGA': '.tga',
	'TARGA_RAW': '.tga',
	'CINEON': '.cin',
	'DPX': '.dpx',
	'OPEN_EXR_MULTILAYER': '.exr',
	'OPEN_EXR': '.exr',
	'HDR': '.hdr',
	'TIFF': '.tif',
	# Add more formats if needed
}

AVAILABLE_FORMATS = ['BMP', 'IRIS', 'PNG', 'JPEG', 'JPEG2000', 'TARGA', 'TARGA_RAW', 'CINEON', 'DPX',
					 'OPEN_EXR_MULTILAYER',
					 'OPEN_EXR', 'HDR', 'TIFF']
"""List of available output file formats"""


def _get_badfname(fname, N=100):
	"""Search for filename in the format
	<main_fname><i:04d>.<ext>
	where i is the frame number. if no file found for i < N, raise error.
	otherwise, return found file
	"""
	f, ext = os.path.splitext(fname)
	for i in range(N):
		fname = f + f"{i:04d}" + ext
		if os.path.isfile(fname):
			return fname

	raise FileNotFoundError(f"File {fname} not found")


def remove_ext(fname):
	return os.path.splitext(fname)[0]  # remove extension if given


class Compositor:
	"""Compositor output - for handling file outputs, and managing Compositor node tree"""

	def __init__(self, view_layer='ViewLayer', background_color: tuple = None,
				 rgb_color_space: str = 'Filmic sRGB'):
		"""
		:param view_layer: Name of View Layer to render
		:param background_color: If given, RGB[A] tuple in range [0-1], will overwrite World background with solid color (while retaining lighting effects).
		:param rgb_color_space: Color transform for RGB only.
		"""
		# Create compositor node tree
		bpy.context.scene.use_nodes = True
		self.node_tree = bpy.context.scene.node_tree

		self.view_layer = view_layer  #

		self.file_output_nodes = {}  # Mapping of output name to FileOutputNode
		self.mask_nodes = {}  # Mapping of mask pass index to CompositorNodeGroup
		self.overlays = {}
		self.aovs = []  # List of AOVs (used to update before rendering)

		# We set view transform to 'Raw' to avoid any gamma correction to all non-Image layers
		bpy.context.scene.view_settings.view_transform = 'Raw'

		# Socket to be used as RGB input for anything. Defined separately in case of applying overlays (e.g. background color)
		self._rgb_socket = get_node_by_name(self.node_tree, 'Render Layers').outputs['Image']
		self._set_rgb_color_space(rgb_color_space)
		if background_color is not None:
			self._set_background_color(background_color)

	def _tidy_tree(self):
		"""Tidy up node tree"""
		tidy_tree(self.node_tree)

	@property
	def render_layers_node(self):
		return get_node_by_name(self.node_tree, 'Render Layers')

	def _get_render_layer_output(self, key: str):
		"""Get output socket from Render Layers node"""
		if key == 'Image':  # special case
			return self._rgb_socket
		else:
			return self.render_layers_node.outputs[key]

	def get_mask(self, index, input_rgb: Union[str, CompositorNodeGroup], anti_aliasing=False) -> CompositorNodeGroup:
		"""Get mask node from pass index. If not found, create new mask node"""
		bpy.context.scene.view_layers[self.view_layer].use_pass_object_index = True  # Make sure object index is enabled

		if index not in self.mask_nodes:

			if isinstance(input_rgb, str):
				ip_node = self._get_render_layer_output(input_rgb)

			elif isinstance(input_rgb, CompositorNodeGroup):
				ip_node = input_rgb.outputs['Image']

			else:
				raise TypeError(f"input_rgb must be str or CompositorNodeGroup, got {type(input_rgb)}")

			dtype = 'Float' if isinstance(ip_node, bpy.types.NodeSocketFloat) else 'Color'

			cng = MaskOverlay(f"Mask - ID: {index} - Input {input_rgb}",
							  self.node_tree, index=index, dtype=dtype,
							  use_antialiasing=anti_aliasing)

			self.node_tree.links.new(ip_node, cng.input('Image'))
			self.node_tree.links.new(self._get_render_layer_output('IndexOB'), cng.input('IndexOB'))
			self.mask_nodes[index] = cng

		self._tidy_tree()
		return self.mask_nodes[index]

	def get_bounding_box_visual(self, col=(0., 0., 255., 255.), thickness:int=5) -> BoundingBoxOverlay:
		"""
		Return a bounding box visual overlay.

		:param col: (3,) or (N, 3) Color(s) of bounding box(es) [in BGR]
		:param thickness: (,) or (N,) Thickness(es) of bounding box(es)
		"""

		cng = BoundingBoxOverlay(f"Bounding Box Visual", self.node_tree, col=col, thickness=thickness)
		self.node_tree.links.new(self._get_render_layer_output('Image'), cng.input('Image'))

		if 'bbox' in self.overlays:
			raise ValueError("Only allowed one BBox overlay (it can contain multiple objects).")

		self.overlays['bbox'] = cng

		self._tidy_tree()
		return cng

	def get_keypoints_visual(self, marker: str = 'x', color: tuple = (0, 0, 255), size: int = 5,
							 thickness: int = 2) -> KeypointsOverlay:
		"""
		Return a keypoints visual overlay.

		:param marker: Marker type, either [c/circle], [s/square], [t/triangle] or [x]. Default 'x'
		:param size: Size of marker. Default 5
		:param color: Color of marker, RGB or RGBA, default (0, 0, 255) (red)
		:param thickness: Thickness of marker. Default 2
		"""

		cng = KeypointsOverlay(f"Keypoints Visual", self.node_tree, marker=marker, color=color,
							   size=size, thickness=thickness)
		self.node_tree.links.new(self._get_render_layer_output('Image'), cng.input('Image'))

		if 'keypoints' in self.overlays:
			raise ValueError("Only allowed one Keypoints overlay.")

		self.overlays['keypoints'] = cng

		self._tidy_tree()
		return cng

	def get_axes_visual(self, size: int = 1, thickness: int = 2) -> AxesOverlay:
		"""
		Return an axes visual overlay.

		:param size: Size of axes. Default 100
		:param thickness: Thickness of axes. Default 2
		"""

		cng = AxesOverlay(f"Axes Visual", self.node_tree,
						  size=size, thickness=thickness)
		self.node_tree.links.new(self._get_render_layer_output('Image'), cng.input('Image'))

		if 'axes' in self.overlays:
			raise ValueError("Only allowed one Axes overlay.")

		self.overlays['axes'] = cng
		self._tidy_tree()
		return cng

	def stack_visuals(self, *visuals: AlphaImageOverlay) -> AlphaImageOverlay:
		"""Given a series of image overlays, stack them and return to be used as a single output node.

		:param *visuals: Stack of overlays to add."""

		if len(visuals) < 2:
			raise ValueError("Requires at least 2 visuals to stack")

		# No need to store these overlays separately in self.overlays, but need to check they're all present
		for overlay in visuals:
			if overlay not in self.overlays.values():
				raise ValueError(
					f"Visual {overlay} not found in Compositor. Make sure it was obtained via the Compositor.")

		# Stack the output of the previous to the input of the next
		for va, vb in zip(visuals, visuals[1:]):
			self.node_tree.links.new(va.output('Image'), vb.input('Image'))

		return visuals[-1]

	def get_depth_visual(self, max_depth=1, col:tuple=(255, 255, 255)) -> CompositorNodeGroup:
		"""Get depth visual, which normalizes depth values so max_depth = col,
		and any values below that are depth/max_depth * col.

		:param max_depth: Maximum depth value to normalize to
		:param col: Color of maximum depth value. 0-255 RGB or RGBA."""

		if 'Depth' not in self.render_layers_node.outputs:
			render_depth()

		# convert col to 0-1, RGBA
		col = ([i / 255 for i in col] + [1])[:4]

		cng = DepthVis(self.node_tree, max_depth=max_depth, col=col)
		self.node_tree.links.new(self._get_render_layer_output('Depth'), cng.input('Depth'))

		self._tidy_tree()
		return cng

	def define_output(self, input_data: Union[str, CompositorNodeGroup, AOV], directory: str = '.',
					  file_name: str = None,
					  file_format: str = 'PNG', color_mode: str = 'RGBA', jpeg_quality: int = 90,
					  png_compression: int = 15, color_depth: str = '8', EXR_color_depth: str = '32',
					  name: str = None) -> str:
		"""Add a connection between a valid render output, and a file output node.

		This should only be called once per output (NOT inside a loop).
		Inside the loop, only call :attr:`~update_filename`, :attr:`update_all_filenames` :attr:`~update_directory`

		All outputs will be defined in raw color space (no color correction), except for the RGB output,
		and any overlays on this output (e.g. Bounding Box Visualization)

		:param input_data: If :class:`str`,  will get the input_data from that key in the render_layers_node. If :class:`~CompositorNodeGroup`, use that node as input. If :class:`AOV`, use that AOV as input (storing AOV).
		:param directory: Directory to save output to
		:param file_name: Name of file to save output to. If None, will use `name` (or `input_data` if `name` is None)
		:param file_format: File format to save output as. Must be in :class:`AVAILABLE_FORMATS`
		:param color_mode: Color mode to save output as.
		:param jpeg_quality: Quality of JPEG output.
		:param png_compression: Compression of PNG output.
		:param color_depth: Color depth of output.
		:param EXR_color_depth: Color depth of EXR output.
		:param name: Name of output. If not given, will take the str representation of input_data

		:returns: Name of output, which can be used to update filename or directory

		"""

		if isinstance(input_data, AOV):
			self.aovs.append(input_data)
			input_data = input_data.name  # name is sufficient to pull from render_layers_node

		assert file_format in format_to_extension, f"File format `{file_format}` not supported. Options are: {list(format_to_extension.keys())}"

		if name is None:
			name = str(input_data)

		node_name = f"File Output {name}"

		# check node doesn't exist
		if name in self.file_output_nodes:
			raise ValueError(f"File output `{name}` already exists. Only call define_output once per output type.")

		if file_name is None:  # if fname is not given, use input_name
			file_name = name

		file_name = remove_ext(file_name)
		directory = os.path.abspath(directory)  # make sure directory is absolute
		os.makedirs(directory, exist_ok=True)

		# Create new 'File Output' node in compositor
		node = self.node_tree.nodes.new('CompositorNodeOutputFile')
		node.name = node_name

		if isinstance(input_data, str):
			self.node_tree.links.new(self._get_render_layer_output(input_data), node.inputs['Image'])

		elif isinstance(input_data, CompositorNodeGroup):  # add overlay in between
			self.node_tree.links.new(input_data.outputs[0], node.inputs['Image'])

		else:
			raise NotImplementedError(
				f"input_data must be either str or CompositorNodeGroup, got {type(input_data)}")

		# Set file output node properties
		node.label = node_name
		node.base_path = directory
		node.file_slots[0].path = file_name

		# File format kwargs
		node.format.file_format = file_format
		node.format.color_mode = color_mode
		node.format.quality = jpeg_quality
		node.format.compression = png_compression
		node.format.color_depth = color_depth if file_format != 'OPEN_EXR' else EXR_color_depth

		self.file_output_nodes[name] = node

		self._tidy_tree()
		return name

	def update_filename(self, key: str, fname: str):
		"""Reassign the filename (not directory) for a given file output node.

		:param key: key of output, as given in `define_output`
		:param fname: new filename, without extension"""
		fname = remove_ext(fname)
		node = self.file_output_nodes[str(key)]
		node.file_slots[0].path = fname

	def update_all_filenames(self, fname: str):
		"""Reassign all filenames (not directories) for all file output nodes.

		:param fname: new filename, without extension"""
		fname = remove_ext(fname)
		for node in self.file_output_nodes.values():
			node.file_slots[0].path = fname

	def update_directory(self, key: str, directory: str):
		"""Reassign the directory for a given file output node

		:param key: key of output, as given in `define_output`
		:param directory: new directory
		"""
		node = self.file_output_nodes[str(key)]
		node.base_path = directory

	def _fix_namings(self, prefix=None):
		"""After rendering,
		File Output node has property that frame number gets added to filename.
		Fix that here"""

		prefix = '' if prefix is None else prefix + '_'

		for node in self.file_output_nodes.values():
			# get expected file name and extension
			ext = format_to_extension[node.format.file_format]
			bad_file_name = _get_badfname(os.path.join(node.base_path, node.file_slots[0].path + ext))
			target_file_name = os.path.join(node.base_path, prefix + node.file_slots[0].path + ext)
			shutil.move(
				bad_file_name,
				target_file_name
			)

	def _update_aovs(self):
		"""Update any AOVs that are connected to the render layers node"""
		for aov in self.aovs:
			aov.update()

	def render(self, camera: Union[Camera, List[Camera]] = None, scene: bpy.types.Scene = None,
			   annotations: AnnotationHandler = None,
			   animation: bool = False, frame_start: int = 0, frame_end: int = 250):
		"""Render the scene.

		:param camera: Camera(s) to render from. If None, will use `scene.camera`. If multiple, will render each camera separately, appending the camera's names as the output file names.
		:param scene: Scene to render. If None, will use `bpy.context.scene`.
		:param annotations: Object containing annotation information for each camera view to be used for overlays
		:param animation: If True, will render an animation, using `frame_start` and `frame_end` as the start and end frames.
		:param frame_start: Start frame for animation.
		:param frame_end: End frame for animation.
		"""

		if scene is None:
			scene = bpy.context.scene

		_original_active_camera = bpy.context.scene.camera
		if camera is None:
			camera = Camera()

		if animation:
			scene.frame_start = frame_start
			scene.frame_end = frame_end

		multi_camera = isinstance(camera, list)
		if not multi_camera:
			camera = [camera]

		for cam in camera:
			if annotations is not None:
				annotation = annotations.get_annotation_by_camera(cam.name)

				# apply overlays PER CAMERA
				for oname, overlay in self.overlays.items():
					overlay.update(getattr(annotation, oname), camera=cam, scene=scene)  # multi kwargs

			self._update_aovs()

			scene.camera = cam.obj
			render(animation=animation)

			if not animation:
				self._fix_namings(prefix=cam.name if multi_camera else None)

		# reset active camera
		scene.camera = _original_active_camera

	def _set_rgb_color_space(self, color_space: str = 'Filmic sRGB'):
		"""Color spaces are all handled manually within compositor (so that we keep
		AOVs in raw space). So set the color space for RGB socket here."""

		color_space_node = self.node_tree.nodes.new('CompositorNodeConvertColorSpace')
		color_space_node.from_color_space = 'Linear'
		color_space_node.to_color_space = color_space

		self.node_tree.links.new(self._rgb_socket, color_space_node.inputs[0])
		self._rgb_socket = color_space_node.outputs[0]

	def _set_background_color(self, color: tuple = (0, 0, 0)):
		"""Set a solid background color, instead of transparent.
		Will remove the visuals of existing world background (but not the lighting effects).

		:param color: RGBA color, in range [0, 1]
		"""

		world.set_transparent()

		rgba = color
		if len(rgba) == 3:
			rgba = (*rgba, 1)

		rgb_node = self.node_tree.nodes.new('CompositorNodeRGB')
		rgb_node.outputs[0].default_value = rgba

		mix_node = self.node_tree.nodes.new('CompositorNodeMixRGB')

		self.node_tree.links.new(self._rgb_socket, mix_node.inputs[2])
		self.node_tree.links.new(rgb_node.outputs[0], mix_node.inputs[1])

		self.node_tree.links.new(self._get_render_layer_output('Alpha'), mix_node.inputs['Fac'])
		self._rgb_socket = mix_node.outputs['Image']



"""Node Group collections for use in Blender"""

import bpy
from ...utils import types
import numpy as np
from mathutils import Vector
from .node_group import ShaderNodeGroup

# docs-special-members: __init__
# no-inherited-members

class DeformedGeneratedTextureCoordinates(ShaderNodeGroup):
	"""
	A :class:`ShaderNodeGroup` that outputs generated texture coordinates, but in the deformed space.

	For example use, see :class:`DisplacementGeneratedAOV <blendersynth.blender.aov.DisplacementGeneratedAOV>`.
	"""

	def __init__(self, node_tree : bpy.types.NodeTree,
		mesh: types.Mesh, bbox_min: types.VectorLike, bbox_max: types.VectorLike):
		"""To calculate the generated space, either needs mesh or bbox_min, bbox_max.

		:param node_tree: NodeTree to add group to
		:param mesh: If given, `bbox_min` and `bbox_max` are calculated using this
		:param bbox_min: If given, `bbox_max` must also be given. The minimum of the bounding box of the mesh
		:param bbox_max: If given, `bbox_min` must also be given. The maximum of the bounding box of the mesh
		"""

		super().__init__("DeformedGeneratedTextureCoordinates", node_tree)

		self.group.outputs.new('NodeSocketVector', 'Vector')

		self.geometry_node = self.add_node('ShaderNodeNewGeometry')
		self.vec_transform_node = self.add_node('ShaderNodeVectorTransform')
		self.map_range_node = self.add_node('ShaderNodeMapRange')

		self.vec_transform_node.vector_type = 'POINT'
		self.vec_transform_node.convert_from = 'WORLD'
		self.vec_transform_node.convert_to = 'OBJECT'

		self.map_range_node.clamp = False # avoid clamping to 0-1
		self.map_range_node.data_type = 'FLOAT_VECTOR'
		self.register_bounds(mesh, bbox_min, bbox_max)

		# link everything up
		self.link(self.geometry_node.outputs['Position'], self.vec_transform_node.inputs['Vector'])
		self.link(self.vec_transform_node.outputs['Vector'], self.map_range_node.inputs['Vector'])
		self.link(self.map_range_node.outputs['Vector'], self.output_node.inputs['Vector'])

		self.tidy()

	def register_bounds(self, mesh: types.Mesh, bbox_min: types.VectorLike, bbox_max: types.VectorLike):
		# calculate bbox_min, bbox_max if not given
		if mesh:
			bbox_min, bbox_max = mesh.get_raw_bounds()

		self.map_range_node.inputs[7].default_value = bbox_min
		self.map_range_node.inputs[8].default_value = bbox_max



import bpy
from bpy_extras.object_utils import world_to_camera_view
import mathutils
import numpy as np


def project_point_to_image(P: mathutils.Vector, scene:bpy.types.Scene,
						   camera:bpy.types.Camera, invert_y:bool=True):
	"""Return 2D (x, y) image coordinates of a 3D point P.

	:param P: 3D point
	:param scene: Blender scene
	:param camera: Camera to project through
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)"""

	ndc_x, ndc_y, ndc_depth = world_to_camera_view(scene, camera, P)  # normalized device coords
	imgx = scene.render.resolution_x * ndc_x

	if invert_y:
		imgy = scene.render.resolution_y * (1 - ndc_y)  # ndc_y measured from bottom
	else:
		imgy = scene.render.resolution_y * ndc_y


	return imgx, imgy


def project_points(points:np.ndarray, scene:bpy.types.Scene, camera:bpy.types.Camera, invert_y:bool=True):
	"""Project points to image plane.

	:param points: Nx3 matrix of points
	:param scene: Blender scene
	:param camera: Camera to project through
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom

	:return: Nx2 matrix of 2D image coordinates"""
	coords_2d = []
	for p in points:
		p_vec = mathutils.Vector(*(p,))
		coords_2d.append(project_point_to_image(p_vec, scene, camera, invert_y=invert_y))

	return np.array(coords_2d)



"""Blender device management."""

import bpy
import warnings
from typing import List


# no-inherited-members

def _isgpu(device):
	return device.type in {'CUDA', 'OPENCL', 'METAL'}


class DeviceList(list):
	@property
	def names(self) -> List[str]:
		"""Return the names of all devices."""
		return [device.name for device in self]


class Devices:
	"""A manager for Blender rendering devices (CPU & GPU)."""

	def __init__(self):
		self.device_list = DeviceList(bpy.context.preferences.addons['cycles'].preferences.devices)

	def set_device_usage(self, cpu: bool = True, cuda: bool = True,
						 opencl: bool = True, metal: bool = True):
		"""Set/unset all devices of certain types to enabled.

		:param cpu: Enable/disable CPU devices
		:param cuda: Enable/disable CUDA devices
		:param opencl: Enable/disable OpenCL devices
		:param metal: Enable/disable Metal devices
		"""

		for device in self.device_list:
			if device.type == 'CPU':
				device.use = cpu

			elif device.type == 'CUDA':
				device.use = cuda

			elif device.type == 'OPENCL':
				device.use = opencl

			elif device.type == 'METAL':
				device.use = metal

			else:
				warnings.warn(f'Unknown device type {device.type}')

	@property
	def available_gpus(self) -> DeviceList:
		"""Return list of all GPU devices."""
		return DeviceList(device for device in self.device_list if _isgpu(device))

	@property
	def enabled_gpus(self) -> DeviceList:
		"""Return list of available GPU devices."""
		return DeviceList(device for device in self.available_gpus if device.use)

	@property
	def device_names(self) -> List[str]:
		"""Return a list of all device names."""
		return self.device_list.names

	@property
	def enabled_device_names(self) -> List[str]:
		"""Return a list of named devices that are enabled."""
		return [device.name for device in self.device_list if device.use]

	def set_by_name(self, name: str, use: bool = True):
		"""Set use/not use based on a device name.

		:param name: Device name
		:param use: Whether to use the device or not
		"""

		for device in self.device_list:
			if device.name == name:
				device.use = use
				return

		raise KeyError(f'No device named {name}. Available devices: {self.device_names}')



import copy

import bpy
from .utils import GetNewObject, animatable_property
from .bsyn_object import BsynObject
from ..utils import types

LIGHT_TYPES = ['POINT', 'SUN', 'SPOT', 'AREA']


class Light(BsynObject):
	"""Light object, for managing lights in the scene"""
	light_types = LIGHT_TYPES
	"""List of available light types"""

	def __init__(self, light):
		self._object = light

	@classmethod
	def from_scene(cls, key):
		"""Create object from scene key

		:param key: Key of object in scene
		:return: :class:`~blendersynth.blender.light.Light` object
		"""
		obj = bpy.data.objects[key]
		return cls(obj)

	@classmethod
	def create(cls, light_type: str = 'POINT', name: str = 'Light', intensity: float = 1.0,
			   color: types.VectorLike = (1.0, 1.0, 1.0), location: types.VectorLike = (0, 0, 0)):
		"""Create a new Light object

		:param light_type: Type of light to create (see :attr:`~blendersynth.blender.light.Light.light_types`)
		:param name: Name of light
		:param intensity: Intensity of light
		:param color: Color of light
		:param location: Location of light

		:return: :class:`~blendersynth.blender.light.Light` object

		"""
		assert light_type in LIGHT_TYPES, f"Light type `{light_type}` not found. Options are: {LIGHT_TYPES}"

		light_data = bpy.data.lights.new(name=name, type=light_type)
		light_data.energy = intensity
		light_data.color = color

		light = bpy.data.objects.new(name=name, object_data=light_data)
		light.location = location
		bpy.context.collection.objects.link(light)

		return Light(light)

	@property
	def energy(self):
		"""Energy of light source"""
		return self.obj.data.energy

	@energy.setter
	def energy(self, value):
		self.set_energy(value)

	@animatable_property('energy', 'data')
	def set_energy(self, value: float):
		"""Set energy of light source

		:param value: Energy of light source
		:return:
		"""
		self.obj.data.energy = value

	@property
	def color(self):
		"""Color of light source"""
		return self.obj.data.color

	@color.setter
	def color(self, value):
		self.set_color(value)

	@animatable_property('color', 'data')
	def set_color(self, value: types.VectorLike):
		"""Set color of light source

		:param value: RGB[A] color
		"""
		self.obj.data.color = value



"""When constructing a dataset, the INPUTS object below will
return an iterable with read's in sys.argv's `--jobs` jsons."""

import sys
import json
import os
import logging
from typing import Union, List

class Inputs:
	"""This class is used to iterate over the JSONs passed in via `--jobs` in sys.argv.

	Will also convert all kwargs to attributes for easy access"""

	jsons = None
	"""List of JSON files passed in via `--jobs` in sys.argv."""

	def __init__(self):
		self.jsons = sys.argv[sys.argv.index('--jobs') + 1].split(',')
		log_loc = sys.argv[sys.argv.index('--log') + 1]

		# Set up logging
		logging.basicConfig(filename=log_loc, level=logging.INFO, filemode='a',
							format='%(asctime)s - %(levelname)s - %(message)s')
		
		# store all --kwargs as attributes
		for k, v in zip(sys.argv, sys.argv[1:]):
			if k.startswith('--'):
				setattr(self, k[2:], v)


	def __iter__(self):
		for n, j in enumerate(self.jsons):
			with open(j, 'r') as f:
				fname = os.path.splitext(os.path.split(j)[-1])[0]
				yield fname, json.load(f)

			# Once we get here, we've passed 'yield', so we know that JSON has been loaded & rendering has occured
			logging.info(f"RENDERED: {fname}")

	def __len__(self):
		return len(self.jsons)

class DebugInputs(Inputs):
	"""Class to emulate the Inputs class, for a single JSON for testing."""
	def __init__(self, json_loc: Union[str, List], repeats:int=1):
		"""

		:param json_loc: Location of JSON(s) to use
		:param repeat: Number of times to repeat the JSON(s)
		"""
		if isinstance(json_loc, str):
			json_loc = [json_loc]

		self.jsons = json_loc * repeats




import bpy

AREA_TYPES = ["VIEW_3D", "IMAGE_EDITOR", "NODE_EDITOR", "SEQUENCE_EDITOR", "CLIP_EDITOR",
			  "DOPESHEET_EDITOR", "GRAPH_EDITOR", "NLA_EDITOR", "TEXT_EDITOR", "CONSOLE",
			  "INFO", "TOPBAR", "STATUSBAR", "OUTLINER", "PROPERTIES", "FILE_BROWSER", "SPREADSHEET", "PREFERENCES"]
"""List of all area types"""

def get_areas(area_type):
	"""Yield all areas of the given type.

	:param area_type: Area type to get (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
	"""
	assert area_type in AREA_TYPES, f"Invalid area type: {area_type}"

	for area in bpy.context.screen.areas:
		if area.type == area_type:
			yield area

def get_area(area_type) -> bpy.types.Area:
	"""Return the first found area of the given type.

	:param area_type: Area type to get (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
	"""
	for area in get_areas(area_type):
		return area

def change_area_to(area_type_from: str, area_type_to: str):
	"""Change the first area found.

	:param area_type_from: Area type to change from (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
	:param area_type_to: Area type to change to (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
	"""
	get_area(area_type_from).type = area_type_to



"""Utilities for neatly arranging nodes in the node editor,
to make debugging easier."""

import bpy
from collections import defaultdict
from copy import copy


def get_source_nodes(node):
	for i in node.inputs:
		for link in i.links:
			yield link.from_node


def get_sink_nodes(node):
	for o in node.outputs:
		for link in o.links:
			yield link.to_node


def split_to_islands(nodes):
	"""Given a node list, return a list of sets, each one being a fully disconnected island"""

	def dfs(node, island):
		"""Depth-first search from node."""
		node['visited'] = True
		islands[node] = island
		for linked_node in [*get_source_nodes(node), *get_sink_nodes(node)]:
			if not linked_node['visited']:
				dfs(linked_node, island)

	islands = {}
	island_counter = 0

	# Set all nodes to unvisited
	for node in nodes:
		node['visited'] = False

	# Start a DFS from each unvisited node
	for node in nodes:
		if not node['visited']:
			dfs(node, island_counter)
			island_counter += 1

	# Group nodes by island
	grouped_islands = []
	for i in range(island_counter):
		grouped_islands.append({k for k, v in islands.items() if v == i})

	return grouped_islands


def calc_depth(node_island):
	"""Given a node island (i.e. all nodes in the same fully connected component),
	set the depth of each node to be the maximum depth of all its input nodes + 1.
	Once complete, normalize these depths so the first item is 0"""

	for nodes in node_island:
		nodes['depth'] = None

	node_island_copy = copy(node_island)
	start_node = node_island_copy.pop()
	start_node['depth'] = 0
	nodes_to_use = {start_node}

	# First Pass - assign a depth to each node by traversing the graph from the start node
	while any(node['depth'] is None for node in node_island):
		node = nodes_to_use.pop()
		for sink_node in get_sink_nodes(node):
			if sink_node in node_island:
				if sink_node['depth'] is None:
					sink_node['depth'] = node['depth'] + 1
					nodes_to_use.add(sink_node)

		for source_node in get_source_nodes(node):
			if source_node in node_island:
				if source_node['depth'] is None:
					source_node['depth'] = node['depth'] - 1
					nodes_to_use.add(source_node)

	# Second Pass - set each depth to be the maximum depth of all its input nodes + 1
	depth_changed = True
	while depth_changed:
		depth_changed = False
		for node in node_island:
			sources = [s for s in get_source_nodes(node) if s in node_island]
			if sources:
				new_depth = max(node['depth'] for node in sources) + 1
				if node['depth'] != new_depth:
					node['depth'] = new_depth
					depth_changed = True

	# Normalize depths
	min_depth = min(node['depth'] for node in node_island)
	for node in node_island:
		node['depth'] -= min_depth

	return node_island


def tidy_tree(node_tree: bpy.types.NodeTree, dX: int = 400, dY: int = 200):
	"""Search through tree, positioning nodes in a grid based on their depth and connectivity.

	:param node_tree: node tree to tidy
	:param dX: horizontal distance between nodes
	:param dY: vertical distance between nodes"""

	nodes = node_tree.nodes
	islands = split_to_islands(nodes)

	y = 0  # track running height to manage multiple islands

	height = defaultdict(int)  # track height of each depth level
	for island in islands:
		island = calc_depth(island)

		# want to center each depth level, so set heights accordingly
		for i in range(max(node['depth'] for node in island) + 1):
			# Adjusting the initial height calculation to account for node sizes.
			depth_nodes = [node for node in island if node['depth'] == i]
			total_height_for_depth = sum(node.dimensions.y for node in depth_nodes)
			spacing_needed = dY * (len(depth_nodes) - 1)
			height[i] = -(total_height_for_depth + spacing_needed) / 2

		for node in island:
			node.location = (node['depth'] * dX, y + height[node['depth']])
			height[node['depth']] += node.dimensions.y + dY  # Incrementing by the node's height

		y += max(height.values()) + dY


"""Import python module - if not installed, install it"""

from .blender_setup.check_blender_install import _install_module
import sys

def import_module(named_module, pip_name = None):
	"""Import python module - if not installed, install it.

	:param named_module: name of module to import (python name)
	:param pip_name: name of module to install (pip name). If None, will use named_module
	"""

	if pip_name is None:
		pip_name = named_module

	try:
		return __import__(named_module)
	except ModuleNotFoundError:
		print(f"Blender executable does not have module `{named_module}` installed. Installing...")
		python_exec = sys.executable
		_install_module(python_exec, pip_name)
		return __import__(named_module)


"""Overlay RGB image in compositor"""
from ..nodes import CompositorNodeGroup
import bpy
import cv2
import numpy as np
from ...file.tempfiles import create_temp_file
from typing import List

class AlphaImageOverlay(CompositorNodeGroup):
	"""Overlay an image on top of the render, using the alpha channel of the image as a mask"""
	def __init__(self, name='AlphaImageOverlay', node_tree=None, scene=None):
		"""Create a mix node which overlays an image on top of the input image."""

		super().__init__(name, node_tree)

		# Set default width and height
		self.width = 1000
		self.height = 1000
		if scene:
			self.width = scene.render.resolution_x
			self.height = scene.render.resolution_y

		# define I/O
		self.group.inputs.new('NodeSocketColor', 'Image')
		self.group.outputs.new('NodeSocketColor', 'Image')

		# create nodes
		self.overlay_img = self.group.nodes.new('CompositorNodeImage')
		self.mix_node = self.group.nodes.new('CompositorNodeMixRGB')
		self.sep_color_node = self.group.nodes.new('CompositorNodeSepRGBA')

		# link up internal nodes
		self.group.links.new(self.overlay_img.outputs['Image'], self.sep_color_node.inputs['Image'])

		self.group.links.new(self.input_node.outputs['Image'], self.mix_node.inputs[1])
		self.group.links.new(self.overlay_img.outputs['Image'], self.mix_node.inputs[2])
		self.group.links.new(self.sep_color_node.outputs[3], self.mix_node.inputs['Fac']) # alpha

		self.group.links.new(self.mix_node.outputs['Image'], self.output_node.inputs['Image'])

		self.create_img()

		self.tidy()

	def create_img(self):

		# create temp image to draw keypoints on, with 8 random alphanumeric characters
		# 8 random alphanum chars
		self.temp_img_loc = create_temp_file('.png')

		# initialize as white
		self.img = np.ones((self.height, self.width, 3), dtype=np.uint8) * 255
		cv2.imwrite(self.temp_img_loc, self.img)

		# connect this image to the overlay node
		self.overlay_img.image = bpy.data.images.load(self.temp_img_loc)
		return self.temp_img_loc


class KeypointsOverlay(AlphaImageOverlay):
	"""Overlay which draws keypoints on top of the render."""
	def __init__(self, name='KeypointsOverlay', node_tree=None, scene=None, camera=None,
				 marker: str = 'x', size: int = 5,
				 color: tuple = (0, 0, 255),
				 thickness:int=2):
		"""

		:param name:
		:param node_tree:
		:param scene:
		:param camera:
		:param marker: Marker type, either [c/circle], [s/square], [t/triangle] or [x]. Default 'x'
		:param size: Size of marker. Default 5
		:param color: Color of marker, RGB or RGBA, default (0, 0, 255) (red)
		:param thickness: Thickness of marker. Default 2
		"""
		super().__init__(name, node_tree)

		self.marker = marker
		self.size = size
		self.color = color
		self.thickness = thickness

	def update(self, keypoints: np.ndarray, scene:bpy.types.Scene=None,
			   camera:bpy.types.Camera=None):
		"""Given [N x 3] keypoints, draw them onto a new temp image.

		:param keypoints: N x 3 keypoints
		:param scene: Scene to draw from
		:param camera: Camera

		"""

		self.width = scene.render.resolution_x
		self.height = scene.render.resolution_y

		# reset image to black, with alpha = 0
		self.img = np.zeros((self.height, self.width, 4), dtype=np.uint8)

		color = self.color
		if len(color) == 3:
			color = (*color, 255) # add alpha

		# draw keypoints on image
		size = self.size
		for kp in keypoints:
			if self.marker in ['c', 'circle']:
				cv2.circle(self.img, (int(kp[0]), int(kp[1])), size, color, self.thickness)
			elif self.marker in ['s', 'square']:
				cv2.rectangle(self.img, (int(kp[0]) - size, int(kp[1]) - size), (int(kp[0]) + size, int(kp[1]) + size),
							  color, self.thickness)
			elif self.marker in ['t', 'triangle']:
				points = np.array([[kp[0], kp[1] - size], [kp[0] - size, kp[1] + size], [kp[0] + size, kp[1] + size]],
								  dtype=np.int32)
				cv2.polylines(self.img, [points], isClosed=True, color=color, thickness=self.thickness)
				cv2.fillPoly(self.img, [points], color=color)
			elif self.marker == 'x':
				cv2.line(self.img, (int(kp[0]) - size, int(kp[1]) - size), (int(kp[0]) + size, int(kp[1]) + size),
						 color, self.thickness)
				cv2.line(self.img, (int(kp[0]) + size, int(kp[1]) - size), (int(kp[0]) - size, int(kp[1]) + size),
						 color, self.thickness)
			else:
				raise ValueError("Invalid marker: {}".format(self.marker))

		self.save_image(self.temp_img_loc, self.img)

class BoundingBoxOverlay(AlphaImageOverlay):
	"""Overlay which draws bounding boxes on top of the render."""
	def __init__(self, name='BoundingBoxOverlay', node_tree=None, scene=None, camera=None,
				 col:tuple=(0, 0, 255, 255), thickness:int=2):
		super().__init__(name, node_tree)
		self.col = col
		self.thickness = int(thickness)  # cv2 requires int

	def update(self, bboxes, scene=None, camera=None):
		"""Given [N x 4] bounding boxes, draw them onto a new temp image."""

		self.width = scene.render.resolution_x
		self.height = scene.render.resolution_y

		# reset image to black, with alpha = 0
		self.img = np.zeros((self.height, self.width, 4), dtype=np.uint8)

		col = self.col
		if len(col) == 3:
			col = (col[0], col[1], col[2], 255)

		# draw bounding boxes on image
		for bbox in bboxes:
			cv2.rectangle(self.img, (int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])),
						  col, self.thickness)

		self.save_image(self.temp_img_loc, self.img)

class AxesOverlay(AlphaImageOverlay):
	"""Overlay which draws axes on top of an existing render"""

	def __init__(self, name='BoundingBoxOverlay', node_tree=None, scene=None, camera=None,
				 size: int = 1, thickness: int = 2):
		super().__init__(name, node_tree)

		self.size = size
		self.thickness = thickness

	def update(self, points:List[np.ndarray], scene:bpy.types.Scene=None,
			   camera:bpy.types.Camera=None):
		"""Plot axes onto the image.

		:param points: list of N x 4 2D points of centre + XYZ unit axes
		:param scene: Scene
		:param camera: Camera"""

		self.width = scene.render.resolution_x
		self.height = scene.render.resolution_y
		self.img = np.zeros((self.height, self.width, 4), dtype=np.uint8)

		for P in points:
			centre = P[0].astype(int)
			for i, point in enumerate(P[1:]):
				offset = point - centre
				point2 = centre + offset * self.size
				x, y = map(int, centre)
				x2, y2 = map(int, point2)
				# note BGR color order
				# and -y because image y-axis is flipped
				self.img = cv2.line(self.img, (x, y), (x2, y2),
							   (255 * (i == 2), 255 * (i == 1), 255 * (i == 0), 255), self.thickness)

		self.save_image(self.temp_img_loc, self.img)


from .bsyn_object import BsynObject
import bpy
from ..utils import types
from .utils import GetNewObject

class Empty(BsynObject):

	def __init__(self, obj: bpy.types.Object = None, name=None, **kwargs):
		if obj is None:
			obj = self._create_empty_in_blender(**kwargs)
		self._object = obj

		if name is not None:
			obj.name = name

	@classmethod
	def create(cls, location: types.VectorLike = None, rotation: types.VectorLike = None,
			   scale: types.VectorLike = None, radius: float = 1.0, name=None, **kwargs):
		"""Create and return a new Empty instance."""
		obj = cls._create_empty_in_blender(location, rotation, scale, radius, **kwargs)
		return cls(obj, name=name)

	@staticmethod
	def _create_empty_in_blender(location: types.VectorLike = None, rotation: types.VectorLike = None,
								 scale: types.VectorLike = None, radius: float = 1.0, **kwargs) -> bpy.types.Object:
		"""Private method to create and return a new Empty object in Blender."""

		importer = GetNewObject(bpy.context.scene)
		kwargs = {**kwargs, **{k: v for k, v in zip(['location', 'rotation', 'scale', 'radius'],
													[location, rotation, scale, radius])}}
		kwargs ={k: v for k, v in kwargs.items() if v is not None}
		with importer:
			bpy.ops.object.empty_add(**kwargs)

		return importer.imported_obj



import bpy
from .utils import project_points
import numpy as np
from .annotation_handler import AnnotationHandler
from ..blender.camera import Camera
from typing import List, Union

def _project_keypoints(points3d: np.ndarray, camera:bpy.types.Camera=None, scene:bpy.types.Scene=None) -> np.ndarray:
	"""Project 3D world points to 2D image coordinates.

	Projects using :func:`blendersynth.annotations.utils.project_points`.

	:param points3d: Nx3 array of 3D points
	:param scene: scene to use (if None, `use bpy.context.scene`)
	:param camera: camera to project through (if None, use `bpy.context.scene.camera`)

	:return: Nx2 array of 2D image coordinates
	"""

	if scene is None:
		scene = bpy.context.scene

	if camera is None:
		camera = scene.camera

	# project points
	return project_points(points3d, scene, camera)

def project_keypoints(points3d: np.ndarray, camera: Union[Camera, List[Camera]] = None, scene: bpy.types.Scene=None) -> AnnotationHandler:
	"""Project 3D world points to 2D image coordinates for each camera.

	:param points3d: Nx3 array of 3D points
	:param camera: Either a single :class:`~blendersynth.blender.camera.Camera` object, or a list of them. If None, use bpy.context.scene.camera
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	"""

	return_dict = {}

	if camera is None:
		camera = bpy.context.scene.camera

	if not isinstance(camera, list):
		camera = [camera]

	for cam in camera:
		return_dict[cam.name] = _project_keypoints(points3d, camera=cam, scene=scene)

	return AnnotationHandler.from_annotations(return_dict, ann_type='keypoints')